#!/usr/bin/env python3

import numpy as np
from os import environ, remove, path
from sys import stderr
import subprocess
import matplotlib.pyplot as plt
from sklearn.cluster import DBSCAN
from sklearn import metrics
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
from json import loads, dump
import codecs
from oxDNA_analysis_tools.output_bonds import output_bonds
from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2, get_input_parameter

#Runs principal component analysis of the points to produce orthogonal variation allowing for flattening of data to 3D
def perform_pca(points, out_dims):
    normed_points = points - np.mean(points, axis=0)
    R = np.cov(normed_points, rowvar=False)
    evals, evecs = np.linalg.eigh(R)
    idx = np.argsort(evals)[::-1]
    evecs = evecs[:,idx]
    evals = evals[idx] 
    evecs = evecs[:, :out_dims]
    return np.dot(evecs.T, normed_points.T).T, evals, evecs

def make_heatmap(inputfile, system, filename):
    """
    Creates a heatmap for each cluster centroid

    Parameters: 
        inputfile (str): The input file used to run the simulation.
        system (base.system): The system to make a heatmap from.
        filename (str): The file to write the heatmap to.
    """
    from contact_map import contact_map
    m = contact_map(inputfile, system, True)
    fig, ax = plt.subplots()
    a = ax.imshow(m, cmap='viridis', origin='lower')
    ax.set(title = "interaction network",
        ylabel="nucleotide id",
        xlabel="nucleotide id")
    b = fig.colorbar(a, ax=ax)
    b.set_label("distance", rotation = 270)
    plt.savefig(filename+".png")


def find_element(n, x, array):
    """
    Finds the id of the nth time element x appears in an array.
    """
    c = 0
    for i, j in enumerate(array):
        if (j == x):
            if (c == n):
                return i
            c += 1
    return -1

def split_trajectory(traj_file, inputfile, labs, n_clusters):
    """
    Splits the trajectory into the clustered trajectories

    Parameters:
        traj_file (str): The analyzed trajectory file.
        inputfile (str): The input file used to run the analyzed simulation.
        labs (numpy.array): The cluster each point belongs to.
    """
    top_file = get_input_parameter(inputfile, "topology")

    print ("cluster\tmembers")

    #energies = []
    #H_counts = []

    for cluster in (set(labs)):
        in_cluster = list(labs).count(cluster)

        print ("{}\t{}".format(cluster, in_cluster))

        #energies.append([])
        #H_counts.append([])

        #for making trajectories of each cluster
        try:
            remove("cluster_"+str(cluster)+".dat")
        except: pass

    confid = 0
    r1 = LorenzoReader2(traj_file, top_file)
    system = r1._get_system() 
    
    print ("INFO: splitting trajectory...", file=stderr)
    print ("INFO: Will write cluster trajectories to cluster_<cluster number>.dat", file=stderr)

    while system != False:
        system.print_traj_output("cluster_"+str(labs[confid])+".dat", "/dev/null")

        ###########
        #If you want to get additional information about a cluster, add that code here
        #for example, if you want average energy and hydrogen bonds:
        '''
        energies[labs[confid]].append(0)
        H_counts[labs[confid]].append(0)
        system.map_nucleotides_to_strands()
        out = output_bonds(inputfile, system)

        for line in out.split('\n'):
            if line[0] != '#' and line[0] != '\n':
                line = line.split(" ")
                for m in line[2:9]:
                    energies[labs[confid]][-1] += float(m)
                if float(line[6]) != 0:
                    H_counts[labs[confid]][-1] += 1
        energies[labs[confid]][-1] /= len(system._nucleotides)
        '''
        ############
            
        confid += 1
        system = r1._get_system()

    #This is where you print the information about each cluster
    '''    
    avg_energy = []
    energy_dev = []
    avg_H = []
    H_dev = []
    for e, h in zip(energies, H_counts):
        avg_energy.append(np.mean(e))
        energy_dev.append(np.std(e))
        avg_H.append(np.mean(h))
        H_dev.append(np.std(h))
    print ("cluster\tmembers\tavg_E\tE_dev\tavg_H\tH_dev")
    for cluster in (set(labs)):
        in_cluster = list(labs).count(cluster)
        print ("{}\t{}\t{:.3f}\t{:.3f}\t{:.3f}\t{:.3f}".format(cluster, in_cluster, avg_energy[cluster], energy_dev[cluster], avg_H[cluster], H_dev[cluster]))
    '''

#Calculates the centroid for each cluster
#Creates a .dat and .top file for each centroid
#also makes a heatmap for each cluster and provides some information
def get_centroid(points, metric_name, num_confs, labs, traj_file, inputfile):
    """
    Takes the output from DBSCAN and produces the trajectory and centroid from each cluster.

    Parameters:
        points (numpy.array): The points fed to the clstering algorithm.
        metric_name (str): The type of data the points represent.
        labs (numpy.array): The cluster each point belongs to.
        traj_file (str): The analyzed trajectory file.
        inputfile (str): The input file used to run the analyzed simulation.
    """
    
    print("INFO: splitting clusters...", file=stderr)
    print("INFO: Will write cluster trajectories to traj_<cluster_number>.dat", file=stderr)
    print ("cluster\tn\tavg_E\tE_dev\tavg_H\tH_dev\tcentroid_t")
    for cluster in (set(labs)):
        if metric_name == "precomputed":
            masked = points[labs == cluster]
            in_cluster_id = np.sum(masked, axis = 1).argmin()

        in_cluster = list(labs).count(cluster)
        centroid_id = find_element(in_cluster_id, cluster, labs)
        top_file = get_input_parameter(inputfile, "topology")

        r = LorenzoReader2(traj_file, top_file)
        output = r._get_system(N_skip=centroid_id)
        filename = "centroid"+str(cluster)

        output.print_lorenzo_output(filename+".dat", filename+".top")
        
        make_heatmap(inputfile, output, filename)

#Runs a DBSCAN on the points matrix and creates a 3D plot showing the clusters
#There is code for both animating the plot and just having an interactive 3D plot.  Comment out the one you don't want
def perform_DBSCAN(points, num_confs, traj_file, inputfile, metric_name):
    """
    Runs the DBSCAN algorithm using the provided analysis as positions and splits the trajectory into clusters.

    Parameters:
        points (numpy.array): The points fed to the clstering algorithm.
        num_confs (int): The number of configurations in the trajectory.
        traj_file (str): The analyzed trajectory file.
        inputfile (str): The input file used to run the analyzed simulation.
        metric_name (str): The type of data the points represent (usually either "euclidean" or "precomputed").
    
    Returns:
        labels (numpy.array): The clusterID of each configuration in the trajectory.
    """

    #run system checks
    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "sklearn", "matplotlib"])
    
    print("INFO: Running DBSCAN...", file=stderr)
    EPS=12
    MIN_SAMPLES=8

    #dump the input as a json file so you can iterate on EPS and MIN_SAMPLES
    dump_file = "cluster_data.json"
    print("INFO: Serializing input data to {}".format(dump_file), file=stderr)
    print("INFO: Run just clustering.py with the serialized data to adjust clustering parameters", file=stderr)
    out = [points.tolist(), num_confs, traj_file, inputfile, metric_name]
    dump(out, codecs.open(dump_file, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)

    #prepping to show the plot later
    #this only shows the first three dimensions because we assume that this is either PCA data or only a few dimensions anyway

    #components = perform_pca(points, 3)
    dimensions = []
    x = []
    dimensions.append(x)

    if points.shape[1] > 1:
        y = []
        dimensions.append(y)

    if points.shape[1] > 2:
        z = []
        dimensions.append(z)
    
    for i in points:
        for j, dim in enumerate(dimensions):
            dim.append(i[j])

    #DBSCAN parameters:
    #eps: the pairwise distance that configurations below are considered neighbors
    #min_samples: The smallest number of neighboring configurations required to start a cluster
    #metric: If the matrix fed in are points in n-dimensional space, then the metric needs to be "euclidean".
    #If the matrix is already a square distance matrix, the metrix needs to be "precomputed".
    #the eps and min_samples need to be determined for each structure
    #If you're making your own multidimensional data, you probably want to normalize your data first.
    print("INFO: Adjust clustering parameters by modifying the 'EPS' and 'MIN_SAMPLES' values in the script.", file=stderr)
    print("INFO: Current values: eps={}, min_samples={}".format(EPS, MIN_SAMPLES))
    db = DBSCAN(eps=EPS, min_samples=MIN_SAMPLES, metric=metric_name).fit(points) 
    labels = db.labels_
    
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    print ("Number of clusters:", n_clusters_)

    
    print("INFO: Making cluster plot...")
    if len(dimensions) == 3:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
    else:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

    plt.xlabel("OP0")
    plt.ylabel("OP1")

    if len(dimensions) == 3:
        ax.set_zlabel("OP2")
        #to show the plot immediatley and interactivley
        '''a = ax.scatter(x, y, z, s=2, alpha=0.4, c=labels, cmap=plt.get_cmap('tab10', 7))
        b = fig.colorbar(a, ax=ax)
        plt.show()'''
        
        #to make a video showing a rotating plot
        plot_file = "animated.mp4"
        def init():
            a = ax.scatter(x, y, z, s=2, alpha=0.4, c=labels, cmap=plt.get_cmap('tab10', n_clusters_+1))
            fig.colorbar(a, ax=ax)
            return [fig]

        def animate(i):
            ax.view_init(elev=10., azim=i)
            return [fig]

        anim = animation.FuncAnimation(fig, animate, init_func=init, frames=range(360), interval=20, blit=True)
        
        anim.save(plot_file, fps=30, extra_args=['-vcodec', 'libx264'])

    else:
        plot_file = "plot.png"
        if len(dimensions) == 1:
            dimensions.append(np.arange(len(dimensions[0])))
            a = ax.scatter(dimensions[1], dimensions[0], s=2, alpha=0.4, c=labels, cmap=plt.get_cmap('tab10', n_clusters_+1))
        else:
            a = ax.scatter(dimensions[0], dimensions[1], s=2, alpha=0.4, c=labels, cmap=plt.get_cmap('tab10', n_clusters_+1))
        b = fig.colorbar(a, ax=ax)
        plt.savefig(plot_file)
    print("INFO: Saved cluster plot to {}".format(plot_file), file=stderr)

    if metric_name == "precomputed":
        get_centroid(points, metric_name, num_confs, labels, traj_file, inputfile)

    split_trajectory(traj_file, inputfile, labels, n_clusters_)

    return labels

def main():
    import argparse
    parser = argparse.ArgumentParser(prog = path.basename(__file__), description="Calculates clusters based on provided order parameters.")
    parser.add_argument('serialized_data', type=str, nargs=1, help="The json-formatted input file")
    args = parser.parse_args()
    data_file = args.serialized_data[0]

    #load a previously serialized dataset
    data = codecs.open(data_file, 'r', encoding='utf-8').read()
    unpacked = loads(data)
    points = np.array(unpacked[0])
    labels = perform_DBSCAN(points, unpacked[1], unpacked[2], unpacked[3], unpacked[4])

if __name__ == '__main__':
    main()
    


#!/usr/bin/env python3
from json import loads
from sys import stderr
import argparse
import numpy as np
import os

def make_dat(mean_info, outfile):
    """
        Converts the json file produced by compute_mean.py into an oxDNA .dat file

        Parameters:
            mean_info (numpy.array): the json file from compute_mean.py parsed as a numpy array.
            outfile (string): the name to save the .dat file to
    """
    mean = np.array(mean_info['g_mean'])
    mmax = mean.max()
    mmin = mean.min()

    #box size is 1.5x the longest dimension of the structure
    box_size = (mmax - mmin).max() * 1.5
    with open(outfile, "w") as file:
        print("INFO: Writing configuration to", outfile, file=stderr)
        file.writelines([
            't = %d\n' % 0,                                       # set time
            'b = %d %d %d\n' % (box_size, box_size, box_size),    # set box size
            'E = 0 0 0\n'])                                       # set energy
        for p, a1, a3 in zip(mean_info["g_mean"], mean_info["a1_mean"] ,mean_info["a3_mean"]):
            file.write(
                "".join([
                        "%f %f %f " % (p[0],p[1],p[2]),
                        "%f %f %f " % (a1[0],a1[1],a1[2]),
                        "%f %f %f " % (a3[0],a3[1],a3[2]),
                        "0 0 0 ",
                        "0 0 0 "
                    ]) + "\n"
            )
def main(): 
    #get commandline arguments 
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Converts a mean structure .json from compute_mean.py to an oxDNA-readable .dat")
    parser.add_argument('mean', type=str, nargs=1, help="A mean structure from compute_mean.py")
    parser.add_argument('output', type=str, nargs=1, help="The name of the output file")
    args = parser.parse_args()

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy"])

    #load the mean file, which is in .json format
    with open(args.mean[0], "r") as file:
        mean_info = loads(
            file.read()
        )

    #write the file out in oxDNA format
    outfile = args.output[0]
    make_dat(mean_info, outfile)

if __name__ == '__main__':
    main()


#!/usr/bin/env python

#Created by: Erik Poppleton
#Date: 6/29/18
#Python2
#Converts the forces file printed out by tiamat2oxdna to a pairs file containing all designed H-bonds

import sys
import argparse
import os

def main():

    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Convert an external force file to a list of particle pairs")
    parser.add_argument('force_file', type=str, nargs=1, help="The force file to generate pairs from")
    parser.add_argument('-o', '--output', type=str, nargs=1, help='name of the file to write the angle list to')

    args = parser.parse_args()

    infile = args.force_file[0]

    #if there's no outfile argument, just print to stdout.
    try:
        out = args.output[0]
        outfile = open(out, 'w+')

    except:
        outfile = sys.stdout

    #Process the forces file
    with open(infile) as f:
        for i, line in enumerate(f.readlines()):
            line = line.strip()
            if line.startswith("particle"):
                a = line.split("=")[1].strip()
            if "ref_particle" in line:
                b = line.split("=")[1].strip()
            if "}" in line:
                if int(a) < int(b):
                    print(a, b, sep = ' ', file=outfile)
                a = -1
                b = -1

if __name__ == '__main__':
    main()
    


import os.path
import sys
import numpy as np
from oxDNA_analysis_tools.UTILS import base
from oxDNA_analysis_tools.UTILS.base_array import base_array

#helper for cal_confs
def blocks(file, size=1000000):
    """
    An iterator that returns chunks of the given file

    Parameters:
        file (file): An open file handler to chunk.
        <optional> size (int): The size of the chunks.  Defaults to 1 MB

    Yields:
        Chunks of the provided file.
    """
    while True:
        b = file.read(size)
        if not b: break
        yield b

#calculates the length of a trajectory file
def cal_confs(traj_file):
    """
    Calculates the number of configurations in a trajectory
    """
    with open(traj_file, "rb") as f:
        return (sum(bl.count(b"t") for bl in blocks(f)))

#gets the value out of an oxDNA input file
def get_input_parameter(input_file, parameter):
    fin = open(input_file)
    value = ''
    for line in fin:
        line = line.lstrip()
        if not line.startswith('#'):
            if parameter in line:
                value = line.split('=')[1].replace(' ','').replace('\n','')
    fin.close()
    if value == '':
        print("ERROR: Key {} not found in input file {}".format(parameter, input_file))
    return value

#The Python3 successor to the trusty LorenzoReader
#Reads in oxDNA trajectory files one configuration at a time to avoid memory overflow
#_get_system() returns a System object as defined in base.py
class LorenzoReader2:
    def __enter__(self):
        return self

    def __exit__(self,  exc_type, exc_val, exc_tb):
        self.__del__()

    def __init__(self, configuration, topology, check_overlap=False):
        # NOTE: on paths with symlinks kills the output stream 
        # so script continues to run ... probably  better to crash 
        if not os.path.isfile(configuration):
            print("Configuration file '{}' is not readable".format(configuration))
        if not os.path.isfile(topology):
            print("Topology file '{}' is not readable".format(topology))
        self._check_overlap = check_overlap
        self._conf = open(configuration, "r")
        with open(topology, "r") as file:
            file.readline()
            self._top_lines = file.readlines()
            self.particle_count = len(self._top_lines)

    def __del__(self):
        try: 
            if self._conf: 
                self._conf.close()
        except:
            print("ERROR: The reader could not load the provided configuration/topology pair")

    def __iter__(self):
        return self

    def __next__(self):
        s = self._get_system()
        if s:
            s.inbox_system()
            return s
        else:
            raise StopIteration
  
    def _read(self, only_strand_ends=False, skip=False):
        timeline = self._conf.readline()
        time = 0
        if  len(timeline) == 0:
            return False
        else:
            time = float(timeline.split()[2])
        
        box = np.array([float(x) for x in self._conf.readline().split()[2:]])
        [E_tot, E_pot, E_kin] = [float(x) for x in self._conf.readline().split()[2:5]]

        if skip: #this could be slightly optimized by moving it up, but then it wouldn't be same length as toplines
            for tl in self._top_lines:
                self._conf.readline()
            return False

        system = base.System(box, time=time, E_pot=E_pot, E_kin=E_kin)
        base.Nucleotide.index = 0
        base.Strand.index = 0

        s = False
        strandid_current = 0
        for tl in self._top_lines:
            tls = tl.split()
            n3 = int(tls[2])
            n5 = int(tls[3])
            strandid = int(tls[0])
            if (len (tls[1]) == 1):
                try:
                    if strandid > 0:
                        b = base.base_to_number[tls[1]]
                    if strandid < 0 :
                        b = base.aa_to_number[tls[1]]
                except:
                    b = 0
                bb = b
            else:
                try:
                    tmp = int (tls[1])
                except:
                    raise ValueError ("problems in topology file with specific base pairing")

                if tmp > 0:
                    b = tmp % 4
                else:
                    b = (3 - ((3 - tmp) % 4))
                bb = tmp

            if strandid != strandid_current:
                # check for circular strand
                if n3 != -1:
                    iscircular = True
                else:
                    iscircular = False

                if s:
                    system.add_strand(s, self._check_overlap)
                if strandid >= 0:
                    s = base.Strand()
                else:
                    s = base.Peptide()
                if iscircular:
                    s.make_circular()
                strandid_current = strandid

            ls = self._conf.readline().split()
            cm = [float(x) for x in ls[0:3]]
            a1 = [float(x) for x in ls[3:6]]
            a3 = [float(x) for x in ls[6:9]]
            v = [float(x) for x in ls[9:12]]
            L = [float(x) for x in ls[12:15]]
            if not only_strand_ends or n3 == -1 or n5 == -1:
                try:
                    s.add_nucleotide(base.Nucleotide(cm, a1, a3, b, bb, v, L, n3))
                except Exception as e:
                    print("ERROR: Reader died while reading configuration with t = {}.\n\nError message:\n{}".format(time, e))
                    return False

        system.add_strand(s, self._check_overlap)

        return system


    # if only_strand_ends == True then a strand will contain only the first and the last nucleotide
    # useful for some analysis like csd for coaxial interactions
    def _get_system(self, only_strand_ends=False, N_skip=0):
        for _ in range(N_skip):
            self._read(skip=True)

        return self._read(only_strand_ends=only_strand_ends, skip=False)


class ErikReader:
    def __enter__(self):
        return self

    def __exit__(self,  exc_type, exc_val, exc_tb):
        self.__del__()

    def __del__(self):
        try: 
            if self._conf: 
                self._conf.close()
        except:
            print("ERROR: The reader could not load the provided configuration")
            sys.exit(1)

    def __init__(self, configuration):
        if not os.path.isfile(configuration):
            print("Configuration file '{}' is not readable".format(configuration))
            sys.exit(1)
        self._conf = open(configuration, "r")
        self._time = None
        self._box = np.zeros(3)
        self._conf_energy = np.zeros(3)
        self._len = 0

    """
    Special reader that handles the first line and allocates the memory for storing the system
    """
    def _read_first(self):
        self._line = self._conf.readline().split()
        if len(self._line) == 0:
            print("ERROR: the configuration file is empty")
        else:
            self._time = float(self._line[2])

        self._line = self._conf.readline().split()
        self._box = np.array(self._line[2:], dtype=float)
        
        self._line = self._conf.readline().split()
        self._conf_energy = np.array(self._line[2:], dtype=float) # keep the energy info (cause it's interesting for remd)

        #we make lists here because we don't know how big the configuration is (that is in the topology)
        positions = []
        a1s = []
        a3s = []
        self._line = self._conf.readline().split()
        while self._line[0] != "t":
            positions.append(self._line[0:3])
            a1s.append(self._line[3:6])
            a3s.append(self._line[6:9])
            self._line = self._conf.readline().split()
            if self._line == []:
                break

        self._configuration = base_array(self._time, self._box, self._conf_energy, np.array(positions, dtype=float), np.array(a1s, dtype=float), np.array(a3s, dtype=float))
        self._len = len(positions)

        return (self._configuration)
    
    """
    Read the next configuration in the trajectory.

    Parameters:
        <optional> n_skip (int): skip the next n configurations before returning

    Returns:
        system (base_array): a set of numpy arrays containing positions and orientations of each nucleotides
    """
    def read(self, n_skip=0):
        if not self._time:
            if n_skip == 0:
                return self._read_first()
            else:
                self._read_first()
                n_skip -= 1

        if n_skip > 0:
            for i in range((n_skip * (self._len + 3)) -1):
                self._conf.readline()
            self._line = self._conf.readline().split() # get the time 

        if  len(self._line) == 0:
            return False
        else:
            self._time = float(self._line[2])
            self._configuration.time = self._time

        self._conf.readline() # we already know the box

        #self._conf.readline() # still don't care about the energy
        self._line = self._conf.readline().split()
        self._conf_energy = np.array(self._line[2:], dtype=float) # keep the energy info (cause it's interesting for remd)
        self._configuration.energy = self._conf_energy

        self._line = self._conf.readline().split() #first line of the configuration
        for i in range(self._len):
            if (len(self._line) == 0 or "t" in self._line) and i < self._len:
                print("ERROR: Reader encountered a partial configuration with only {} lines ({} expected)".format(i, self._len))
            self._configuration.positions[i] = np.array(self._line[0:3], dtype=float)
            self._configuration.a1s[i] = np.array(self._line[3:6], dtype=float)
            self._configuration.a3s[i] = np.array(self._line[6:9], dtype=float)
            self._line = self._conf.readline().split()
        return (self._configuration)
    



        



#!/usr/bin/env python

#Created by: Erik Poppleton
#Date: before 6/21/18
#Python3
#Calculates a vector corresponding to every duplex in a structure.  
#The output from this file is used by duplex_angle_finder.py to make plots of structure flexibility.

from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2, cal_confs, get_input_parameter
import numpy as np
import argparse
from os import environ, path
from sys import stderr
from oxDNA_analysis_tools.UTILS import geom
from oxDNA_analysis_tools.output_bonds import output_bonds
from oxDNA_analysis_tools.UTILS import parallelize_lorenzo_onefile

class Duplex:
    """
        Defines a nucleic acid duplex structure

        Init Parameters:
            index (int): Unique identifier for the duplex
            start_s1 (int): Particle ID of the first particle in the first strand
            end_s1 (int): Particle ID of the last particle in the first strand
            start_s2 (int): Particle ID of the first particle in the complementary strand
            end_s2 (int): Particle ID of the last particle in the complementary strand

        Member Functions:
            get_len()
    """
    def __init__(self, index, start_s1, end_s1, start_s2, end_s2):
        self.index = index
        self.start1 = start_s1
        self.end1 = end_s1
        self.start2 = start_s2
        self.end2 = end_s2

    def get_len(self):
        """
            Returns the length of the duplex
        """
        return self.end1-self.start1

def find_duplex(system):
    """
        Searches a system that has had its hydrogen bonds mapped for duplexes

        Parameters:
            system (base.System): The system to search for duplexes

        Returns:
            duplex_list (list): A list of Duplex objects found in the system
    """
    #create a list of base-paired bases
    paired_bases = []
    for s in system._strands:
        for n in s._nucleotides:
            #print(n.index, n.interactions)
            if len(n.interactions) == 1:
                paired_bases.append(np.array([int(n.index), int(n.interactions[0])]))

    #create a list of duplexes                                             
    duplex_index = 0
    search_start = 0
    duplex_list = []
    while search_start < len(paired_bases)-1:
        start1 = paired_bases[search_start][0]
        end2 = paired_bases[search_start][1]
        if end2 > start1:
            duplex_length = 1
            start_address = search_start
            while start1 + duplex_length == paired_bases[start_address+duplex_length][0] and end2 - duplex_length == paired_bases[start_address+duplex_length][1] and end2-duplex_length > start1+duplex_length:
                duplex_length+=1
                search_start+=1

            if duplex_length > 3:
                duplex_list.append(Duplex(duplex_index, start1,  paired_bases[start_address+duplex_length-1][0], paired_bases[start_address+duplex_length-1][1], end2))
                duplex_index+=1
            search_start+=1
 
        else:
            search_start+=1       

    return duplex_list

def find_angles(reader, inputfile, num_confs, start=None, stop=None):
    """
        Fits a vector to every duplex in each snapshot of a trajectory

        Parameters:
            reader (readers.LorenzoReader2): An active reader on the trajectory file in which to find duplexes.
            num_confs (int): The number of configurations in the reader.  
            <optional> start (int): The starting configuration ID to begin averaging at.  Used if parallel.
            <optional> stop (int): The configuration ID on which to end the averaging.  Used if parallel.

        Returns:
            duplexes_at_step (list): A list containing information about every duplex found in the trajectory.
    """
    if stop is None:
        stop = num_confs
    else: stop = int(stop)
    if start is None:
        start = 0
    else: start = int(start)
    confid = 0
    mysystem = reader._get_system(N_skip = start)
                                                                       
    #launches DNAnalysis and prints out a list of all bonds
    duplexes_at_step = []
    while mysystem != False and confid < stop:
        print("Working on  t = {}".format(mysystem._time))
        mysystem.map_nucleotides_to_strands()
        out = output_bonds(inputfile, mysystem)
        if out == '':
            duplex_list = []
            duplexes_at_step.append(duplex_list)
            confid += 1 
            mysystem = reader._get_system()
            continue
        mysystem.read_H_bonds_output_bonds(out)
        duplex_list = find_duplex(mysystem)

        #call geom.get_axis on each duplex
        for i in range(0, len(duplex_list)):
            #print("start1: ", duplex_list[i].start1, "end2: ", duplex_list[i].end2, "end1: ", duplex_list[i].end1, "start2: ", duplex_list[i].start2)
            if environ.get("OXRNA") == "1":
                duplex_list[i].axis, duplex_list[i].final_hel_pos = geom.get_RNA_axis(mysystem, duplex_list[i].start1, duplex_list[i].end1, duplex_list[i].end2, duplex_list[i].start2, only_plane_vector = True)
            else:
                duplex_list[i].axis, duplex_list[i].final_hel_pos = geom.get_DNA_axis(mysystem, duplex_list[i].start1, duplex_list[i].end1, duplex_list[i].end2, duplex_list[i].start2, only_plane_vector = True)
            #print(duplex_list[i].start1, duplex_list[i].end1, duplex_list[i].end2, duplex_list[i].start2)
            duplex_list[i].time = mysystem._time
        duplexes_at_step.append(duplex_list)
        confid += 1 
        mysystem = reader._get_system()

        #for line in duplex_list:
        #    print("[{}, {}, {}, {}, {}, {}],".format(line.final_hel_pos[0], line.final_hel_pos[1], line.final_hel_pos[2], line.axis[0], line.axis[1], line.axis[2]))
    
    return(duplexes_at_step)

def main():
    parser = argparse.ArgumentParser(prog = path.basename(__file__), description="Fit vectors to every duplex in the structure")
    parser.add_argument('-p', metavar='num_cpus', nargs=1, type=int, dest='parallel', help="(optional) How many cores to use")
    parser.add_argument('inputfile', type=str, nargs=1, help="The inputfile used to run the simulation")
    parser.add_argument('trajectory', type=str, nargs=1, help="The trajectory file from the simulation")
    parser.add_argument('-o', '--output', metavar='output_file',  type=str, nargs=1, help='name of the file to write the angle list to')
    args = parser.parse_args()

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy"])

    #Process command line arguments:
    inputfile = args.inputfile[0]
    traj_file = args.trajectory[0]
    parallel = args.parallel
    if parallel:
        n_cpus = args.parallel[0]

    #-o names the output file
    if args.output:
        outfile = args.output[0]
    else: 
        outfile = "angles.txt"
        print("INFO: No outfile name provided, defaulting to \"{}\"".format(outfile), file=stderr)
        

    #Get relevant parameters from the input file
    top_file = get_input_parameter(inputfile, "topology")
    if "RNA" in get_input_parameter(inputfile, "interaction_type"):
        environ["OXRNA"] = "1"
    else:
        environ["OXRNA"] = "0"

    #Calculate the number of configurations.
    num_confs = cal_confs(traj_file)

    r0 = LorenzoReader2(traj_file, top_file)
    r0._get_system()
    
    #launch find_angle using the appropriate number of threads to find all duplexes.
    if not parallel:
        print("INFO: Fitting duplexes to {} configurations using 1 core.".format(num_confs), file=stderr)
        r = LorenzoReader2(traj_file,top_file)
        duplexes_at_step = find_angles(r, inputfile, num_confs)

    if parallel:
        print("INFO: Fitting duplexes to {} configurations using {} cores.".format(num_confs, n_cpus), file=stderr)
        duplexes_at_step = []
        out = parallelize_lorenzo_onefile.fire_multiprocess(traj_file, top_file, find_angles, num_confs, n_cpus, inputfile)
        [duplexes_at_step.extend(i) for i in out]
    
    if [] in duplexes_at_step:
        print("WARNING: Some configurations were invalid and not included in the analysis.  Please check the log to view the error", file=stderr)

    #print duplexes to a file
    print("INFO: Writing duplex data to {}.  Use duplex_angle_plotter to graph data".format(outfile), file=stderr)
    output = open(outfile, 'w')
    output.write("time\tduplex\tstart1\tend1\tstart2\tend2\taxisX\taxisY\taxisZ\thel_pos\n")
    for i in range (0, len(duplexes_at_step)):
        for j in range(0, len(duplexes_at_step[i])):
            line = '{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t[{},{},{}]\n'.format(duplexes_at_step[i][j].time,duplexes_at_step[i][j].index,duplexes_at_step[i][j].start1,duplexes_at_step[i][j].end1,duplexes_at_step[i][j].start2,duplexes_at_step[i][j].end2,duplexes_at_step[i][j].axis[0],duplexes_at_step[i][j].axis[1],duplexes_at_step[i][j].axis[2],duplexes_at_step[i][j].final_hel_pos[0],duplexes_at_step[i][j].final_hel_pos[1],duplexes_at_step[i][j].final_hel_pos[2])
            output.write(line)
    output.close()

if __name__ == '__main__':
    main()
    



#!/usr/bin/env python3
try:
    from Bio.SVDSuperimposer import SVDSuperimposer
except:
    from bio.SVDSuperimposer import SVDSuperimposer
import numpy as np
from json import loads, dumps
from sys import exit, stderr, argv
from oxDNA_analysis_tools.UTILS.readers import ErikReader, cal_confs
from random import randint
import argparse
import time
import os

from oxDNA_analysis_tools.compute_deviations import compute_deviations
start_t = time.time()

def pick_starting_configuration(traj_file, max_bound, align=None):
    """
        Pick a random conf out of the trajectory file to use as the reference structure.
        
        We assume that that is optimal to align against.  Based on experience, the choice of reference configuration has very little impact on the mean structure

        Parameters: 
            traj_file (string): The name of the trajectory file
            max_bound (int): The reference configuration will be chosen at random from the first max_bound configurations in the trajectory file

        Returns:
            stop_at (int): The configuration ID of the reference configuration
            initial_structure (base.System): The oxDNA system representing the reference configuration.
    """
    with ErikReader(traj_file) as reader:
        if align:
            stop_at = int(align)
        else:
            stop_at = randint(0, max_bound-1)
        print("INFO: We chose {} as reference".format(stop_at), file=stderr)
        initial_structure = reader.read(n_skip=stop_at)
        if not initial_structure:
            print("ERROR: Couldn't read structure at conf num {0}.".format(stop_at), file=stderr)
            exit(1)
        print("INFO: reference structure loaded", file=stderr)
        initial_structure.inbox()
    return stop_at, initial_structure

def normalize(v):
    """
        Return a normalized copy of vector v

        Parameters: 
            v (numpy.array): the vector to be normalized.

        Returns: 
            v / norm(v) (numpy.array).
    """
    norm = np.linalg.norm(v)
    if norm == 0:
        print("WARNING: tried to normalize a vector with 0 length", file=stderr)
        return v
    return v / norm

# helper to prepare a configuration of np.array coordinates
# into smth json is able to serialize
prep_pos_for_json = lambda conf: list(
                        list(p) for p in conf
                    )

def compute_mean (reader, align_conf, indexes, num_confs, start = None, stop = None):
    """
        Computes the mean structure of a trajectory

        Structured to work with the multiprocessing process from UTILS/parallelize.py

        Parameters:
            reader (readers.ErikReader): An active reader on the trajectory file to take the mean of.
            align_conf (numpy.array): The position of each particle in the reference configuration.  A 3xN array.
            num_confs (int): The number of configurations in the reader.  
            <optional> start (int): The starting configuration ID to begin averaging at.  Used if parallel.
            <optional> stop (int): The configuration ID on which to end the averaging.  Used if parallel.
        
        Returns:
            mean_pos_storage (numpy.array): For each particle, the sum of positions in all configurations read.
            mean_a1_storage (numpy.array): For each particle, the sum of a1 orientation vectors in all configuraitons read.
            mean_a3_storage (numpy.array): For each particle, the sum of a3 orientation vectors in all configuraitons read.
            intermediate_mean_structures (list): mean structures computed periodically during the summing to check decoorrelation.
            confid (int): the number of configurations summed for the storage arrays.
    """
    parallel = True
    if stop is None:
        stop = num_confs
    else: stop = int(stop)
    if start is None:
        start = 0
        parallel = False
    else: start = int(start)

    mysystem = reader.read(n_skip = start)
    n_nuc = len(mysystem.positions)

    if not parallel:
        #This also computes the mean every num_confs/10 configurations to check decorrelation.
        #Only works when run in serial.
        INTERMEDIATE_EVERY = np.floor(num_confs / 10)
        # storage for the intermediate mean structures

    intermediate_mean_structures = []

    # the class doing the alignment of 2 structures
    sup = SVDSuperimposer()

    mean_pos_storage = np.zeros((n_nuc, 3))
    mean_a1_storage = np.zeros((n_nuc, 3))
    mean_a3_storage = np.zeros((n_nuc, 3))

    # for every conf in the current trajectory we calculate the global mean
    confid = 0

    while mysystem != False and confid < stop:
        mysystem.inbox()
        indexed_cur_conf_pos = mysystem.positions[indexes]

        # calculate alignment
        sup.set(align_conf, indexed_cur_conf_pos)
        sup.run()
        rot, tran = sup.get_rotran()

        #apply alignment
        mysystem.positions = np.einsum('ij, ki -> kj', rot, mysystem.positions) + tran
        mysystem.a1s = np.einsum('ij, ki -> kj', rot, mysystem.a1s)
        mysystem.a3s = np.einsum('ij, ki -> kj', rot, mysystem.a3s)
        mean_pos_storage += mysystem.positions
        mean_a1_storage += mysystem.a1s
        mean_a3_storage += mysystem.a3s

        # print the rmsd of the alignment in case anyone is interested...
        print("Frame:", confid, "Time:", mysystem.time, "RMSF:", sup.get_rms())
        # thats all we do for a frame
        confid += 1
        mysystem = reader.read()

        # We produce 10 intermediate means to check decorrelation.
        # This can't be done neatly in parallel
        if not parallel and confid % INTERMEDIATE_EVERY == 0:
            mp = np.copy(mean_pos_storage)
            mp /= confid
            intermediate_mean_structures.append(
                prep_pos_for_json(mp)
            )
            print("INFO: Calculated intermediate mean for {} ".format(confid))

    return(mean_pos_storage, mean_a1_storage, mean_a3_storage, intermediate_mean_structures, confid)



def main():
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Computes the mean structure of a trajectory file")
    parser.add_argument('trajectory', type=str, nargs=1, help='the trajectory file you wish to analyze')
    parser.add_argument('-p', metavar='num_cpus', nargs=1, type=int, dest='parallel', help="(optional) How many cores to use")
    parser.add_argument('-o', '--output', metavar='output_file', nargs=1, help='The filename to save the mean structure to')
    parser.add_argument('-f', '--format', metavar='<json/oxDNA/both>', nargs=1, help='Output format for the mean file.  Defaults to json.  Options are \"json\", \"oxdna/oxDNA\", and \"both\"')
    parser.add_argument('-d', '--deviations', metavar='deviation_file', nargs=1, help='Immediatley run compute_deviations.py from the output')
    parser.add_argument('-i', metavar='index_file', dest='index_file', nargs=1, help='Compute mean structure of a subset of particles from a space-separated list in the provided file')
    parser.add_argument('-a', '--align', metavar='alignment_configuration', nargs=1, help='The id of the configuration to align to, otherwise random')
    args = parser.parse_args()

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "Bio", "numpy"])

    #get file names
    traj_file = args.trajectory[0]
    parallel = args.parallel
    if parallel:
        from oxDNA_analysis_tools.UTILS import parallelize_erik_onefile
        n_cpus = args.parallel[0]

    #-f defines the format of the output file
    outjson = False
    outoxdna = False
    if args.format:
        if "json" in args.format:
            outjson = True
        if "oxDNA" in args.format or "oxdna" in args.format:
            outoxdna = True
        if "both" in args.format:
            outjson = True
            outoxdna = True
        if outjson == outoxdna == False:
            print("ERROR: unrecognized output format\nAccepted formats are \"json\", \"oxDNA/oxdna\", and \"both\"", file=stderr)
            exit(1)
    else:
        print("INFO: No output format specified, defaulting to oxDNA", file=stderr)
        outoxdna = True

    #-o names the output file
    if args.output:
        outfile = args.output[0]
    else: 
        if outjson and not outoxdna:
            ext = ".json"
        elif outjson and outoxdna:
            ext = ".json/.dat"
        elif outoxdna and not outjson:
            ext = ".dat"
        outfile = "mean{}".format(ext)
        print("INFO: No outfile name provided, defaulting to \"{}\"".format(outfile), file=stderr)

    #-d will run compute_deviations.py when this script is completed.
    dev_file = None
    if args.deviations:
        dev_file = args.deviations[0]

    #-i will make it only run on a subset of nucleotides.
    #The index file is a space-separated list of particle IDs
    if args.index_file:
        index_file = args.index_file[0]
        with open(index_file, 'r') as f:
            indexes = f.readline().split()
            try:
                indexes = [int(i) for i in indexes]
            except:
                print("ERROR: The index file must be a space-seperated list of particles.  These can be generated using oxView by clicking the \"Download Selected Base List\" button")
    else: 
        with ErikReader(traj_file) as r:
            indexes = list(range(len(r.read().positions)))


    # The reference configuration which is used to define alignment
    align_conf = []

    #calculate the number of configurations in the trajectory 
    num_confs = cal_confs(traj_file)

    # if we have no align_conf we need to chose one
    # and realign its cms to be @ 0,0,0
    if align_conf == []:
        align = None
        if args.align:
            align = args.align[0]
        align_conf_id, align_poses = pick_starting_configuration(traj_file, num_confs, align)
        # we are just interested in the nucleotide positions
        align_conf = align_poses.positions[indexes]

    #Actually compute mean structure
    if not parallel:
        print("INFO: Computing mean of {} configurations with an alignment of {} particles using 1 core.".format(num_confs, len(align_conf)), file=stderr)
        r = ErikReader(traj_file)
        mean_pos_storage, mean_a1_storage, mean_a3_storage, intermediate_mean_structures, processed_frames = compute_mean(r, align_conf, indexes, num_confs)

    #If parallel, the trajectory is split into a number of chunks equal to the number of CPUs available.
    #Each of those chunks is then calculated seperatley and the result is summed.
    if parallel:
        print("INFO: Computing mean of {} configurations with an alignment of {} particles using {} cores.".format(num_confs, len(align_conf), n_cpus), file=stderr)
        out = parallelize_erik_onefile.fire_multiprocess(traj_file, compute_mean, num_confs, n_cpus, align_conf, indexes)
        mean_pos_storage = np.sum(np.array([i[0] for i in out]), axis=0)
        mean_a1_storage = np.sum(np.array([i[1] for i in out]), axis=0)
        mean_a3_storage = np.sum(np.array([i[2] for i in out]), axis=0)
        intermediate_mean_structures = []
        [intermediate_mean_structures.extend(i[3]) for i in out]
        processed_frames = sum((i[4] for i in out))
    # finished task entry
    print("INFO: processed frames total: {}".format(processed_frames), file=stderr)

    #Convert mean structure to a json file
    mean_file = dumps({
                "i_means" : intermediate_mean_structures,
                "g_mean" : prep_pos_for_json(
                    mean_pos_storage / processed_frames
                ),
                "a1_mean" : prep_pos_for_json(
                   [normalize(v)  for v in (mean_a1_storage / processed_frames)]
                ),
                "a3_mean" : prep_pos_for_json(
                   [normalize(v)  for v in (mean_a3_storage / processed_frames)]
                ),
                "p_frames" : processed_frames,
                "ini_conf":{
                    "conf": prep_pos_for_json(align_conf),
                    "id"  : align_conf_id
                }
            })

    #Save the mean structure to the specified output file.
    if outjson or dev_file:
        #save output as json format
        if outoxdna == True:
            #if making both outputs, automatically set file extensions.
            jsonfile = outfile.split(".")[0]+".json"
        else:
            jsonfile = outfile
        print("INFO: Writing mean configuration to", jsonfile, file=stderr)
        with open(jsonfile, "w") as file:
            file.write(mean_file)
    
    if outoxdna:
        #save output as oxDNA .dat format
        if outjson == True:
            #if making both outputs, automatically set file extensions.
            outname = outfile.split(".")[0]+".dat"
        else:
            outname = outfile
        from oxDNA_analysis_tools.mean2dat import make_dat
        make_dat(loads(mean_file), outname)

    #If requested, run compute_deviations.py using the output from this script.
    if dev_file:
        print("INFO: launching compute_deviations.py", file=stderr)

        #this is probably horrible practice, but to maintain the ability to call things from the command line, I cannot pass arguments between main() calls.
        #so instead we're gonna spoof a global variable to make it look like compute_deviations was called explicitally
        argv.clear()
        argv.extend(['compute_deviations.py', '-o', dev_file, "-r", dev_file.split('.')[0]+"_rmsd.png", "-d", dev_file.split('.')[0]+"_rmsd_data.json"])
        if args.index_file:
            argv.append("-i")
            argv.append(index_file)
        if parallel:
            argv.append("-p") 
            argv.append(str(n_cpus))
        argv.append(jsonfile)
        argv.append(traj_file)

        from oxDNA_analysis_tools import compute_deviations
        from sys import executable
        print(executable)
        print(argv)

        compute_deviations.main()

        #compute_deviations needs the json meanfile, but its not useful for visualization
        #so we dump it
        if not outjson:
            print("INFO: deleting {}".format(jsonfile), file=stderr)
            from os import remove
            remove(jsonfile)

    print(time.time() - start_t)

if __name__ == '__main__':
    main()
    


#!/usr/bin/env python3

#PCA.py
#Written by: Erik Poppleton
#Date: 3/6/2019
#Performs a principal component analysis on a trajectory
#the output JSON can be loaded into the viewer where it will overlay as arrows

import numpy as np 
from oxDNA_analysis_tools.UTILS.readers import ErikReader, cal_confs, get_input_parameter
from sys import exit, stderr
import argparse
from json import load, dumps
try:
    from Bio.SVDSuperimposer import SVDSuperimposer
except:
    from bio.SVDSuperimposer import SVDSuperimposer
from random import randint
from oxDNA_analysis_tools.UTILS import parallelize_erik_onefile
from warnings import catch_warnings, simplefilter
from os import environ, path
from oxDNA_analysis_tools.config import check_dependencies

def make_heatmap(covariance):
    """
        Produces a heatmat plot of the covariance between every particle

        Parameters:
            covariance (numpy.array): The covariance matrix of particle positions

        Displays:
            A matplotlib imshow plot of the covariance
    """
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    a = ax.imshow(covariance, cmap='viridis', origin='lower')
    ax.set(title = "interaction network",
       ylabel="nucleotide id*3",
       xlabel="nucleotide id*3")
    b = fig.colorbar(a, ax=ax)
    b.set_label("covariance", rotation = 270)
    plt.savefig("heatmap.png")

def get_cov(reader, align_conf, num_confs, start=None, stop=None):
    """
        Performs principal component analysis on deviations from the mean structure

        Parameters:
            reader (readers.ErikReader): An active reader on the trajectory file to analyze.
            align_conf (numpy.array): The position of each particle in the mean configuration.  A 3xN array.
            num_confs (int): The number of configurations in the reader.  
            <optional> start (int): The starting configuration ID to begin averaging at.  Used if parallel.
            <optional> stop (int): The configuration ID on which to end the averaging.  Used if parallel.

        Returns:
            deviations_marix (numpy.array): The difference in position from the mean for each configuration.
    """
    if stop is None:
        stop = num_confs
    else: stop = int(stop)
    if start is None:
        start = 0
    else: start = int(start)
    
    mysystem = reader.read(n_skip = start)
    
    covariation_matrix = np.zeros((len(mysystem.positions)*3, len(mysystem.positions)*3))
    sup = SVDSuperimposer()
    confid = 0

    #for every configuration in the trajectory chunk, align it to the mean and compute positional difference for every particle
    while mysystem != False and confid < stop:
        print("-->", "frame", confid, "time={}".format(mysystem.time))
        mysystem.inbox()
        cur_conf = mysystem.positions
        sup.set(align_conf, cur_conf)
        sup.run()
        rot, tran = sup.get_rotran()
        #equivalent to taking the dot product of the rotation array and every vector in the deviations array
        cur_conf = np.einsum('ij, ki -> kj', rot, cur_conf) + tran
        difference_matrix = (cur_conf - align_conf).flatten()
        covariation_matrix += np.einsum('i,j -> ij', difference_matrix, difference_matrix)

        confid += 1
        mysystem = reader.read()

    return covariation_matrix

def change_basis(reader, align_conf, components, num_confs, start=None, stop=None):
    """
    Transforms each configuration in a trajectory into a point in principal component space

    Parameters:
        reader (readers.ErikReader): An active reader on the trajectory file to analyze.
        align_conf (numpy.array): The position of each particle in the mean configuration.  A 3xN array.
        components (numpy.array): The principal components of the trajectory.  A 3*Nx3*N array.
        num_confs (int): The number of configurations in the reader.  
        <optional> start (int): The starting configuration ID to begin averaging at.  Used if parallel.
        <optional> stop (int): The configuration ID on which to end the averaging.  Used if parallel.

    Returns:
        coordinates (numpy.array): The positions of each frame of the trajectory in principal component space.
    """

    if stop is None:
        stop = num_confs
    else: stop = int(stop)
    if start is None:
        start = 0
    else: start = int(start)
    
    mysystem = reader.read(n_skip = start)

    coordinates = np.empty((stop, len(mysystem.positions)*3))
    coordinates2 = np.empty((stop, len(mysystem.positions)*3))
    sup = SVDSuperimposer()
    confid = 0

    while mysystem != False and confid < stop:
        print("-->", "frame", confid, "time={}".format(mysystem.time))
        mysystem.inbox()
        cur_conf = mysystem.positions
        sup.set(align_conf, cur_conf)
        sup.run()
        rot, tran = sup.get_rotran()
        #equivalent to taking the dot product of the rotation array and every vector in the deviations array
        cur_conf = np.einsum('ij, ki -> kj', rot, cur_conf) + tran
        coordinates[confid] = np.dot(components, cur_conf.flatten())

        confid += 1
        mysystem = reader.read()

    return(coordinates)

def main():
    parser = argparse.ArgumentParser(prog = path.basename(__file__), description="Calculates a principal component analysis of nucleotide deviations over a trajectory")
    parser.add_argument('inputfile', type=str, nargs=1, help="The inputfile used to run the simulation")
    parser.add_argument('trajectory', type=str, nargs=1, help='the trajectory file you wish to analyze')
    parser.add_argument('meanfile', type=str, nargs=1, help='The mean structure .json file from compute_mean.py')
    parser.add_argument('outfile', type=str, nargs=1, help='the name of the .json file where the PCA will be written')
    parser.add_argument('-p', metavar='num_cpus', nargs=1, type=int, dest='parallel', help="(optional) How many cores to use")    
    parser.add_argument('-c', metavar='cluster', dest='cluster', action='store_const', const=True, default=False, help="Run the clusterer on each configuration's position in PCA space?")
    args = parser.parse_args()

    check_dependencies(["python", "numpy", "Bio"])

    traj_file = args.trajectory[0]
    inputfile = args.inputfile[0] 
    mean_file = args.meanfile[0]
    outfile = args.outfile[0]
    parallel = args.parallel
    if parallel:
        n_cpus = args.parallel[0]
    #-c makes it run the clusterer on the output
    cluster = args.cluster

    num_confs = cal_confs(traj_file)
    
    if mean_file.split(".")[-1] == "json":
        with open(mean_file) as file:
            align_conf = load(file)['g_mean']

    elif mean_file.split(".")[-1] == "dat" or mean_file.split(".")[-1] == "conf" or mean_file.split(".")[-1] == "oxdna":
        with ErikReader(mean_file) as reader:
            align_conf = reader.read().positions
    else:
        print("ERROR: {} is an unrecognized file type. \nThe mean structure must either be provided as an oxDNA configuration file with the extension .dat, .conf or .oxdna or as the .json file produced by compute_mean.py.", file=stderr)
        exit(1)

    cms = np.mean(align_conf, axis=0) #all structures must have the same center of mass
    align_conf -= cms 
        
    #Compute the deviations
    if not parallel:
        r = ErikReader(traj_file)
        covariation_matrix = get_cov(r, align_conf, num_confs)
    
    if parallel:
        out = parallelize_erik_onefile.fire_multiprocess(traj_file, get_cov, num_confs, n_cpus, align_conf)
        covariation_matrix = np.sum([i for i in out], axis=0)

    covariation_matrix /= (num_confs-1)

    #now that we have the covatiation matrix we're going to use eigendecomposition to get the principal components.
    #make_heatmap(covariance)
    print("INFO: calculating eigenvectors", file=stderr)
    evalues, evectors = np.linalg.eig(covariation_matrix) #these eigenvalues are already sorted
    evectors = evectors.T #vectors come out as the columns of the array
    print("INFO: eigenvectors calculated", file=stderr)
    
    import matplotlib.pyplot as plt
    print("INFO: Saving scree plot to scree.png", file=stderr)
    plt.scatter(range(0, len(evalues)), evalues, s=25)
    plt.xlabel("component")
    plt.ylabel("eigenvalue")
    plt.savefig("scree.png")

    total = sum(evalues)
    running = 0
    i = 0
    while running < 0.9:
        running += (evalues[i] / total)
        i += 1
    
    print("90% of the variance is found in the first {} components".format(i))


    
    #if you want to weight the components by their eigenvectors
    #mul = np.einsum('ij,i->ij',evectors, evalues)
    mul = evectors

    #reconstruct configurations in component space
    #because we donlist't save the difference matrix, this involves running through the whole trajectory again
    if not parallel:
        r = ErikReader(traj_file)
        coordinates = change_basis(r, align_conf, mul, num_confs)
    if parallel:
        out = parallelize_erik_onefile.fire_multiprocess(traj_file, change_basis, num_confs, n_cpus, align_conf, mul)
        coordinates = np.concatenate([i for i in out])

    #make a quick plot from the first three components
    print("INFO: Creating coordinate plot from first three eigenvectors.  Saving to coordinates.png", file=stderr)
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.scatter(coordinates[:,0], coordinates[:,1], coordinates[:,2], c='g', s=25)
    plt.savefig("coordinates.png")
    
    #Create an oxView overlays for the first N components
    N = 3
    prep_pos_for_json = lambda conf: list(
                        list(p) for p in conf
                    )
    print("INFO: Change the number of eigenvalues to sum and display by modifying the N variable in the script.  Current value: {}".format(N), file=stderr)
    for i in range(0, N): #how many eigenvalues do you want?
        try:
            if outfile.split(".")[1] != "json":
                raise Exception
            f = outfile.split(".")[0] + str(i) + "." + outfile.split(".")[1]
        except:
            print("ERROR: oxView overlays must have a '.json' extension.  No overlays will be produced", file=stderr)
            break
        out = np.sqrt(evalues[i])*evectors[i]

        with catch_warnings(): #this produces an annoying warning about casting complex values to real values that is not relevant
            simplefilter("ignore")
            output_vectors = out.reshape(int(out.shape[0]/3), 3).astype(float)
        
        with open(f, "w+") as file:
            file.write(dumps({
                "pca" : prep_pos_for_json(output_vectors)
            }))

    #If we're running clustering, feed the linear terms into the clusterer
    if cluster:
        print("INFO: Mapping configurations to component space...", file=stderr)

        #If you want to cluster on only some of the components, uncomment this
        #out = out[:,0:3]

        from clustering import perform_DBSCAN
        labs = perform_DBSCAN(coordinates, num_confs, traj_file, inputfile, "euclidean")

if __name__ == '__main__':
    main()    




#This script should not assume that center of the helix is on the line between two bonded bases
#Created by: Petr Sulc
import oxDNA_analysis_tools.UTILS.base
from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2
try:
    import numpy as np
except:
    import mynumpy as np
import os.path
import sys
import math

prune = 1


def my_norm(vec):
	return np.sqrt(np.dot(vec,vec))

def get_turn_per_bp(s,nucid,hel_vector):
    #this function returns the distance between midpoints of hb sites
    box = s._box
    i1A = nucid
    i1B = len(s._strands[1]._nucleotides)-1-nucid
    firstA = s._strands[0]._nucleotides[i1A]
    firstB = s._strands[1]._nucleotides[i1B] 	
    secondA = s._strands[0]._nucleotides[i1A+1]
    secondB = s._strands[1]._nucleotides[i1B-1]
    first_midpos = (firstA.get_pos_back() - firstB.get_pos_back())   
    second_midpos = (secondA.get_pos_back() - secondB.get_pos_back())   
    first_midpos /= math.sqrt(np.dot(first_midpos,first_midpos))
    second_midpos /= math.sqrt(np.dot(second_midpos,second_midpos))
    first_midpos = first_midpos - np.dot(hel_vector,first_midpos) * hel_vector
    second_midpos = second_midpos - np.dot(hel_vector,second_midpos) * hel_vector
    first_midpos /= math.sqrt(np.dot(first_midpos,first_midpos))
    second_midpos /= math.sqrt(np.dot(second_midpos,second_midpos))
 
    #angle = math.acos(np.dot(first_midpos,hel_vector)) - math.acos(np.dot(second_midpos,hel_vector))
    angle = math.acos(np.dot(first_midpos,second_midpos))
    return angle 

def fit_plane(points):
	#fits plane through points, return normal to the plane:
	rc= np.array(np.zeros(3))
	A = np.array([np.zeros(3),np.zeros(3),np.zeros(3)])
	for point in points:
		rc += point
	rc /= len(points)
	#print(rc)

	for point in points:
		point  = point - rc
		for i in range(3):
			for j in range(3):
				A[i][j] += point[i]*point[j]

	#this could be made faster by using a symmetric eigensolver
	vals, vecs = np.linalg.eigh(A) 		
	#print vals, vecs		
	return vecs[:,0]

def get_local_axis(s,base_id,local_length=4):
	#i = base_id
	posBs = []
	posAs = []
	vectorsA = []
	back_poses = []
	for i in range(base_id,base_id+local_length):	
		nucA = s._nucleotides[i]
		nucB= s._nucleotides[s._N - i-1]
		nucAc = s._nucleotides[i+1]
		nucBc = s._nucleotides[s._N - i- 1 -1]
		midpoint_A = 0.5*(nucA.get_pos_stack() + nucB.get_pos_stack())	
		midpoint_B = 0.5*(nucBc.get_pos_stack() + nucAc.get_pos_stack())
		vectorsA.append(-midpoint_A + midpoint_B)
		posAs.append(nucA.get_pos_back())
		posBs.append(nucB.get_pos_back())
		posAs.append(nucAc.get_pos_back())
		posBs.append(nucBc.get_pos_back())

		back_poses.append(nucA.get_pos_back() - nucAc.get_pos_back() )
		back_poses.append(nucB.get_pos_back() - nucBc.get_pos_back() )
		back_poses.append(midpoint_A - midpoint_B)	
		guess =  (-midpoint_A + midpoint_B) #this is the vector pointing from midpoint of the first bp to the last bp, an estimate of the vector
		guess = guess / my_norm(guess)
		plane_vector = fit_plane(back_poses)
		
		if(np.dot(guess,plane_vector) < 0):
			#print 'Debug Warning, plane vector pointing in opposite direction'
			plane_vector = -1. * plane_vector
	hel_pos = []
	
	#project to the plane
	apos = posAs[0]
	bpos = posBs[0]
	apos = apos - np.dot(apos,plane_vector) * plane_vector
	bpos = bpos - np.dot(bpos,plane_vector) * plane_vector
	bp_vec = -apos + bpos
	midpointA = 0.5 * (apos + bpos)
	perpendicular_vecA = np.cross(bp_vec,plane_vector)
	perpendicular_vecA /= my_norm(perpendicular_vecA)
	apos = posAs[1]
	bpos = posBs[1]
	apos = apos - np.dot(apos,plane_vector) * plane_vector
	bpos = bpos - np.dot(bpos,plane_vector) * plane_vector
	bp_vec = -apos + bpos
	midpointB = 0.5 * (apos + bpos)
	perpendicular_vecB = np.cross(bp_vec,plane_vector)
	perpendicular_vecB /= my_norm(perpendicular_vecB)
	mat = np.array([perpendicular_vecA, -perpendicular_vecB ]).transpose()
	y = midpointB - midpointA
	t,c = np.linalg.lstsq(mat,y)[0]
	if my_norm(midpointA + t*perpendicular_vecA  -  (midpointB + c *perpendicular_vecB)) > 1.e-6 :
		print('Debug Error in finding common intersection point',midpointA + t*perpendicular_vecA , mipointB + c *perpendicular_vecB)
	hel_position = midpointA + t*perpendicular_vecA	
	hel_pos.append(hel_position)	

	print('Position of bp',i,' is ',hel_position)
		
	final_hel_pos = np.zeros(3)
	for pos in hel_pos:
		final_hel_pos += pos
	final_hel_pos /= len(hel_pos)	
					

	return plane_vector,final_hel_pos


def get_inclination(s,plane_vector,i):
	posa = s._nucleotides[i].get_pos_back()
	posb = s._nucleotides[s._N - 1 - i].get_pos_back()
	a1a = s._nucleotides[i]._a1
	a1b = -s._nucleotides[s._N - 1 - i]._a1

	inclination_vector = posa - posb
	inclination_vector /= my_norm(inclination_vector)
	angle = np.rad2deg(math.acos( np.dot(inclination_vector,plane_vector ) ))
	angleA = np.rad2deg(math.acos( np.dot(a1a,plane_vector ) ))
	angleB = np.rad2deg(math.acos( np.dot(a1b,plane_vector ) ))
	angle = 0.5*(angleA + angleB)
	return angle

	
def get_data_with_local_axis(s,first_base=0,last_base=-1,only_plane_vector=False):
	if last_base == -1:
		last_base = s._N/2 - 1
	i = first_base
	nucA = s._nucleotides[i]
	nucB= s._nucleotides[s._N - i-1]
	midA = (nucA.get_pos_back() + nucB.get_pos_back() ) / 2.
	i = last_base
	nucAc = s._nucleotides[i]
	nucBc = s._nucleotides[s._N - i -1]
	midAc = (nucAc.get_pos_back() + nucBc.get_pos_back() ) / 2.
	guess =  (-midA + midAc) #this is the vector pointing from midpoint of the first bp to the last bp, an estimate of the vector
	guess = guess / my_norm(guess)

	vectorsA = []
	posAs = []
	posBs = []
	back_poses = []	
	for i in range(first_base,last_base):
		nucA = s._nucleotides[i]
		nucB= s._nucleotides[s._N - i-1]
		nucAc = s._nucleotides[i+1]
		nucBc = s._nucleotides[s._N - i- 1 -1]
		midpoint_A = 0.5*(nucA.get_pos_back() + nucB.get_pos_back())	
		midpoint_B = 0.5*(nucBc.get_pos_back() + nucAc.get_pos_back())
		vectorsA.append(-midpoint_A + midpoint_B)
		posAs.append(nucA.get_pos_back())
		posBs.append(nucB.get_pos_back())

		back_poses.append(nucA.get_pos_back() - nucAc.get_pos_back() )
		back_poses.append(nucB.get_pos_back() - nucBc.get_pos_back() )
		
		if(i == last_base - 1):
			posAs.append(nucAc.get_pos_back())
			posBs.append(nucBc.get_pos_back())
	
	#now we find the point where the helical vecotr originates

	#now, we can do some statistics: inclination, angle_per_bp, width, rise_per_bp
	inclinations = []
	angles = []
	widths = []
	rises = []
	dislocations = []

	for i in range(len(posAs)-1):
		#print 'Debug, len is ',len(posAs), 'i is',i
		plane_vector,final_hel_pos = get_local_axis(s,i)
		#print 'Debug: local plane vector is : ',plane_vector
		vec_bp = -posAs[i] + posBs[i]
		vec_bp_norm = -vec_bp / my_norm(vec_bp)	
		inclinations.append(90-get_inclination(s,plane_vector,i))
		#inclinations.append(90. - np.rad2deg(math.acos(np.dot(vec_bp_norm,plane_vector))))
		
		#now lift the plane so that it is in the plane of the bp
		midpointA = 0.5 * (posAs[i] + posBs[i])
		midpointB = 0.5 * (posAs[i+1] + posBs[i+1])
		dA = np.dot(plane_vector,midpointA)
		dB = np.dot(plane_vector,midpointB)
		rises.append(dB - dA)
			
		
		midpointA_projection = midpointA - plane_vector * np.dot(midpointA,plane_vector)
		midpointB_projection = midpointB - plane_vector * np.dot(midpointB,plane_vector)
		avec = midpointA_projection - final_hel_pos
		bvec = midpointB_projection - final_hel_pos
		dislocations.append(my_norm(avec))
		dislocations.append(my_norm(bvec))
		#avec = avec / my_norm(avec)
		#bvec = bvec /  my_norm(bvec)
		#angle = np.rad2deg(math.acos(np.dot(avec,bvec)))
		#vec_bp_2 = -posAs[i+1] + posBs[i+1]
		#vec_bp_2 /= my_norm(vec_bp_2)
		#angle = 180. - np.rad2deg(math.acos(np.dot(vec_bp_norm,vec_bp_2)))
		vec_bp_2 = -posAs[i+1] + posBs[i+1]
		vec_bp_2 /= my_norm(vec_bp_2)
		#do projection to the helical axis
		vec_bp_norm = vec_bp_norm - plane_vector * np.dot(vec_bp_norm,plane_vector)
		vec_bp_2 = vec_bp_2 - plane_vector * np.dot(vec_bp_2,plane_vector)
		vec_bp_2 /= my_norm(vec_bp_2)
		vec_bp_norm /= my_norm(vec_bp_norm)
		
		angle = 180. - np.rad2deg(math.acos(np.dot(vec_bp_norm,vec_bp_2)))
		 
		#angle = np.rad2deg(get_turn_per_bp(s,i,plane_vector))
		angles.append(angle)
		
		posA_projection = posAs[i] - plane_vector * np.dot(posAs[i],plane_vector)
		posB_projection = posBs[i] - plane_vector * np.dot(posBs[i],plane_vector)
		wA = my_norm(posA_projection - final_hel_pos)
		wB = my_norm(posB_projection - final_hel_pos)
		widths.append(wA)
		widths.append(wB)	

	return plane_vector, final_hel_pos, inclinations, angles, widths, rises, dislocations	

#needs to be modified so it can handle strand id and first and last base in duplex
def get_RNA_axis(s,first_base, last_base, cfirst_base, clast_base, only_plane_vector=False):

	nucA = s._nucleotides[first_base]
	nucB= s._nucleotides[cfirst_base]
	midA = (nucA.get_pos_back() + nucB.get_pos_back() ) / 2.
	nucAc = s._nucleotides[last_base]
	nucBc = s._nucleotides[clast_base]
	midAc = (nucAc.get_pos_back() + nucBc.get_pos_back() ) / 2.
	guess =  (-midA + midAc) #this is the vector pointing from midpoint of the first bp to the last bp, an estimate of the vector
	guess = guess / my_norm(guess)

	vectorsA = []
	posAs = []
	posBs = []
	back_poses = []	
	
	for i in range(0, last_base-first_base):
		nucA = s._nucleotides[first_base+i]
		nucB= s._nucleotides[cfirst_base-i]
		nucAc = s._nucleotides[first_base+i+1]
		#print(nucA.index, nucAc.index)
		nucBc = s._nucleotides[cfirst_base-i-1]
		midpoint_A = 0.5*(nucA.get_pos_back() + nucB.get_pos_back())	
		midpoint_B = 0.5*(nucBc.get_pos_back() + nucAc.get_pos_back())
		vectorsA.append(-midpoint_A + midpoint_B)
		posAs.append(nucA.get_pos_back())
		posBs.append(nucB.get_pos_back())
		if(i == last_base - 1):
			posAs.append(nucAc.get_pos_back())
			posBs.append(nucBc.get_pos_back())
		back_poses.append(-nucA.get_pos_back() + nucAc.get_pos_back() )
		back_poses.append(-nucB.get_pos_back() + nucBc.get_pos_back() )

	#now we try to fit a plane through all these points
	plane_vector = fit_plane(back_poses)
	#print(plane_vector)
	#plane_vector = fit_plane(vectorsA)

	if(np.dot(guess/my_norm(guess),plane_vector) < 0):
		#print 'Warning, plane vector pointing in opposite direction'
		plane_vector = -1. * plane_vector
	if(np.rad2deg(math.acos(np.dot(plane_vector,guess))) > 20):
		print ('Warning, guess vector and plane vector have angles:', np.rad2deg(math.acos(np.dot(guess/my_norm(guess),plane_vector))))

	#now we find the point where the helical vecotr originates
	hel_pos = []
	for i in range(len(posAs)-1):
		#project to the plane
		apos = posAs[i]
		bpos = posBs[i]
		apos = apos - np.dot(apos,plane_vector) * plane_vector
		bpos = bpos - np.dot(bpos,plane_vector) * plane_vector
		bp_vec = -apos + bpos
		if np.linalg.norm(bp_vec) == 0:
			    continue
		midpointA = 0.5 * (apos + bpos)
		perpendicular_vecA = np.cross(bp_vec,plane_vector)
		perpendicular_vecA /= my_norm(perpendicular_vecA)
		apos = posAs[i+1]
		bpos = posBs[i+1]
		apos = apos - np.dot(apos,plane_vector) * plane_vector
		bpos = bpos - np.dot(bpos,plane_vector) * plane_vector
		bp_vec = -apos + bpos
		if np.linalg.norm(bp_vec) == 0:
			continue
		midpointB = 0.5 * (apos + bpos)
		perpendicular_vecB = np.cross(bp_vec,plane_vector)
		perpendicular_vecB /= my_norm(perpendicular_vecB)
		mat = np.array([perpendicular_vecA, -perpendicular_vecB ]).transpose()
		y = midpointB - midpointA
		t,c = np.linalg.lstsq(mat,y)[0]
		if my_norm(midpointA + t*perpendicular_vecA  -  (midpointB + c *perpendicular_vecB)) > 1.e-6 :
			print ('Error in finding common intersection point',midpointA + t*perpendicular_vecA , mipointB + c *perpendicular_vecB)
		hel_position = midpointA + t*perpendicular_vecA	
		#print ('Hel position from BP',i,'is ',hel_position)
		hel_pos.append(hel_position)	
				
	final_hel_pos = np.zeros(3)
	for pos in hel_pos:
		final_hel_pos += pos
	final_hel_pos /= len(hel_pos)				

	if only_plane_vector:
		return plane_vector, final_hel_pos#, back_poses, guess 
		
	#now, we can do some statistics: inclination, angle_per_bp, width, rise_per_bp
	inclinations = []
	angles = []
	widths = []
	rises = []
	dislocations = []

	for i in range(len(posAs)-1):
		vec_bp = -posAs[i] + posBs[i]
		vec_bp_norm = -vec_bp / my_norm(vec_bp)	
		
		inclinations.append(90-get_inclination(s,plane_vector,i))
		#inclinations.append(90. - np.rad2deg(math.acos(np.dot(vec_bp_norm,plane_vector))))
		
		#now lift the plane so that it is in the plane of the bp
		midpointA = 0.5 * (posAs[i] + posBs[i])
		midpointB = 0.5 * (posAs[i+1] + posBs[i+1])
		dA = np.dot(plane_vector,midpointA)
		dB = np.dot(plane_vector,midpointB)
		rises.append(dB - dA)
		#print 'MidpointA', midpointA, 'MidpointB', midpointB, 'Difference', dB - dA
			
		hel_pointA = final_hel_pos + dA * plane_vector
		hel_pointB = final_hel_pos + dB * plane_vector	
		
		midpointA_projection = midpointA - plane_vector * np.dot(midpointA,plane_vector)
		midpointB_projection = midpointB - plane_vector * np.dot(midpointB,plane_vector)
		avec = midpointA_projection - final_hel_pos
		bvec = midpointB_projection - final_hel_pos
		dislocations.append(my_norm(avec))
		dislocations.append(my_norm(bvec))
		#avec = avec / my_norm(avec)
		#bvec = bvec /  my_norm(bvec)
		#angle =  np.rad2deg(math.acos(np.dot(avec,bvec)))
		#angle =  np.rad2deg(math.acos(np.dot(avec,bvec)))

		vec_bp_2 = -posAs[i+1] + posBs[i+1]
		vec_bp_2 /= my_norm(vec_bp_2)
		#do projection to the helical axis
		vec_bp_norm = vec_bp_norm - plane_vector * np.dot(vec_bp_norm,plane_vector)
		vec_bp_2 = vec_bp_2 - plane_vector * np.dot(vec_bp_2,plane_vector)
		vec_bp_2 /= my_norm(vec_bp_2)
		vec_bp_norm /= my_norm(vec_bp_norm)
		
		angle = 180. - np.rad2deg(math.acos(np.dot(vec_bp_norm,vec_bp_2)))
		 
		#angle = np.rad2deg(get_turn_per_bp(s,i,plane_vector))
		angles.append(angle)
		
		posA_projection = posAs[i] - plane_vector * np.dot(posAs[i],plane_vector)
		posB_projection = posBs[i] - plane_vector * np.dot(posBs[i],plane_vector)
		wA = my_norm(posA_projection - final_hel_pos)
		wB = my_norm(posB_projection - final_hel_pos)
		widths.append(wA)
		widths.append(wB)	

	return plane_vector, final_hel_pos, inclinations, angles, widths, rises, dislocations	

def get_DNA_axis (s,first_base, last_base, cfirst_base, clast_base, only_plane_vector=False):
	vec = np.empty((last_base-first_base, 3))
	for i in range(last_base-first_base):
		nucA = s._nucleotides[first_base+i]
		nucAc= s._nucleotides[cfirst_base-i]
		midA = (nucA.get_pos_base() + nucAc.get_pos_base() ) / 2.
		nucB = s._nucleotides[first_base+i+1]
		nucBc = s._nucleotides[cfirst_base-i-1] 
		midB = (nucB.get_pos_base() + nucBc.get_pos_base()) / 2.
		vec[i] = (midB - midA)/(np.linalg.norm(midB - midA))

	#print(s._nucleotides[cfirst_base].index, s._nucleotides[cfirst_base].get_pos_base())

	pos = (s._nucleotides[first_base].get_pos_base() + s._nucleotides[cfirst_base].get_pos_base() ) / 2.
	vector = np.mean(vec, axis=0)
	return(vector, pos)

#-----------------

def get_bb_dist(s,nucid): 
    #this function measures the distance between backbone sites of the opposite nucleotides
    box = s._box
    i1A = nucid
    i1B = len(s._strands[1]._nucleotides)-1-nucid
    firstA = s._strands[0]._nucleotides[i1A]
    firstB = s._strands[1]._nucleotides[i1B] 	
    first_midpos = (firstA.get_pos_back() - firstB.get_pos_back())   
    return math.sqrt(np.dot(first_midpos,first_midpos)) 

def get_back_back_distance(s,nucid):
    box = s._box
    i1A = nucid
    i1B = len(s._strands[1]._nucleotides)-1-nucid
    firstA = s._strands[0]._nucleotides[i1A]
    firstB = s._strands[1]._nucleotides[i1B] 	
    secondA = s._strands[0]._nucleotides[i1A+1]
    secondB = s._strands[1]._nucleotides[i1B-1]
    first_midpos = (firstA.get_pos_back() - secondA.get_pos_back())   
    second_midpos = (firstB.get_pos_back() - secondB.get_pos_back())
    return math.sqrt( np.dot(first_midpos,first_midpos)), math.sqrt(np.dot(second_midpos,second_midpos))

def get_end_j(s,nid1,nid2):
    i1A = nid1
    i2A = nid2
    i1B = len(s._strands[1]._nucleotides) - nid1 - 1
    i2B = len(s._strands[1]._nucleotides) - nid2 - 1 
    
    firstA = s._strands[0]._nucleotides[i1A]
    firstB = s._strands[1]._nucleotides[i1B] 	
    lastA = s._strands[0]._nucleotides[i2A]
    lastB = s._strands[1]._nucleotides[i2B]
   
    first_midpos = (firstA.get_pos_base() + firstB.get_pos_base()) / 2.0  
    last_midpos = (lastA.get_pos_base() + lastB.get_pos_base()) / 2.0  
  
    box = s._box  
    r0N = last_midpos - first_midpos 
    r0N -= box * np.rint (r0N / box)
    return math.sqrt(np.dot(r0N,r0N)) / float((nid2 - nid1))



'''
if len(sys.argv) < 3:
    base.Logger.log("Usage is %s configuration topology [offset=4] [prune=1]" % sys.argv[0], base.Logger.CRITICAL)
    sys.exit()

l = readers.LorenzoReader(sys.argv[1], sys.argv[2])
s = l.get_system()

offset = 4
prune = 1
if len(sys.argv) >= 4:
    offset = int(sys.argv[3])

if len(sys.argv) >= 5:
    prune = int(sys.argv[4])
    


i1A = offset
i1B = len(s._strands[1]._nucleotides) - offset - 1

print  "#Nucleotides", i1A, i1B


L2 = 0.
l0 = 0.
Ll0 = 0.
niter = 0
rises = []
read_confs = 1
end_rises = []
angles = []
helix_widths = []
incsA = []
incsB = []

bbAs = []
bbBs = []


all_inclinations = []
all_widths = []
all_rises = []
all_dislocations = []
all_angles = []

read_confs = 0

while s:
    if(read_confs % prune != 0):
	read_confs += 1
    	s = l.get_system()
	continue     
    end_rises.append(get_end_j(s,i1A,i1B))
 
    box = s._box  
    plane_vector, final_hel_pos, inclinations, angles, widths, rises, dislocations = get_axis(s,first_base=offset, last_base=len(s._strands[0]._nucleotides)-offset-2,only_plane_vector=False)
    all_inclinations = all_inclinations + inclinations
    all_angles = all_angles + angles
    all_widths = all_widths + widths
    all_rises = all_rises + rises
    all_dislocations = all_dislocations + dislocations
    #print dislocations
    print angles
    print inclinations
    print '#Plane vector:',plane_vector
    print '#Position of helix',final_hel_pos
    for j in range (offset,len(s._strands[0]._nucleotides) - offset - 2):
	jA = j
	
	bbA,bbB = get_back_back_distance(s,jA)
	bbAs.append(bbA)
	bbBs.append(bbB)

    s = l.get_system()
    niter += 1


print '# ',sys.argv[1]
print '# read configurations: ', niter
print 'Rise per bp: %f (+/- %f), Rise (from end-end-dist): %f, Angle (per bp): %f (+/- %f) = pitch %f (+/- %f), width %f (+/- %f), inclination : %f (+- %f), back-back distances:A %f (+- %f), B: %f (+- %f), dislocations: %f (+- %f) \n' % ( np.mean(all_rises)  ,np.std(all_rises) ,np.mean(end_rises),np.mean(all_angles),np.std(all_angles),360. / np.mean(all_angles) , np.std(360. / np.array(all_angles)), np.mean(all_widths), np.std(all_widths) ,(np.mean(all_inclinations)),(np.std(all_inclinations)),  np.mean(bbAs),np.std(bbAs),np.mean(bbBs),np.std(bbBs), np.mean(all_dislocations), np.std(all_dislocations))
'''




#!/usr/bin/env python

from __future__ import print_function
from os import path
from sys import exit, stderr

#set the path to your compiled copy of DNAnalysis here
def set_analysis_path():
	PROCESSPROGRAM = '/home/erik/software/oxDNA/build/bin/DNAnalysis'

	if not path.isfile(PROCESSPROGRAM):
		print ("ERROR: Cannot execute DNAnalysis program. Please edit config.py to point to your compiled DNAnalysis. Current target:", PROCESSPROGRAM, file=stderr)
		print()
		print("config.py can be found in", path.realpath(__file__), file=stderr)
		exit()
	return PROCESSPROGRAM

#checking dependencies to make sure everything is correct
def check_dependencies(to_check):
	flag = False
	dependencies = {
		"numpy": 1.14,
		"matplotlib": 3.0,
		"Bio": 1.73,
		"sklearn": 0.21,
		"pathos": 0.2,
	}
	real_names = {
		"numpy": "Numpy",
		"matplotlib": "MatPlotLib",
		"Bio": "BioPython",
		"sklearn": "SciKit-Learn",
		"pathos": "Pathos"
	}
	websites = {
		"numpy": "numpy.org", 
		"matplotlib": "matplotlib.org",
		"Bio": "biopython.org",
		"sklearn": "scikit-learn.org",
		"pathos": "pypi.org/project/pathos/"
	}

	#check python version
	if "python" in to_check:
		from sys import version_info
		ver = '.'.join([str(i) for i in version_info[0:2]])
		print("INFO: Python version: {}".format('.'.join([str(i) for i in version_info[0:3]])), file=stderr)
		if version_info < (3, 6):
			flag = True
			print("WARNING: Some scripts will not run with Python versions earler than 3.6.  You have {}, please update your environment", file=stderr)

	#check packages
	for package in to_check:
		if package == "python": continue
		try:
			mod = __import__(package)
			print("INFO: Package {} found. Version: {}".format(real_names[package], mod.__version__), file=stderr)
		except:
			if package == "Bio":
				try:
					mod = __import__('bio')
					print("INFO: Package {} found. Version: {}".format(real_names[package], mod.__version__), file=stderr)
				except:
					flag = True
					print("ERROR: Unable to find package {}.  Please check your environment or follow the installation instructions at {}".format(real_names[package], websites[package]), file=stderr)
					continue
			else:
				flag = True
				print("ERROR: Unable to find package {}.  Please check your environment or follow the installation instructions at {}".format(real_names[package], websites[package]), file=stderr)
				continue
		ver = float('.'.join(mod.__version__.split(".")[0:2]))
		if ver < dependencies[package]:
			flag = True
			print("WARNING: Your version for package {} is {}.  This tool was tested using {}.  You may need to update your environment".format(real_names[package], ver, dependencies[package]), file=stderr)

	if flag:
		print("WARNING: Some packages need to be installed/updated.", file=stderr)
	else:
		print("INFO: No dependency issues found.", file=stderr)

	return flag

def main():
	check_dependencies(["python", "numpy", "matplotlib", "Bio", "sklearn", "pathos"])
	p = set_analysis_path()
	print("INFO: DNAnalysis found at:", p, file=stderr)

if __name__ == '__main__':
    main()
    



"""
This parallelizer splits a trajectory into multiple tempfiles and attaches a LorenzoReader to each one.
This is more memory-intensive than one-file, but is a significant performance increase on certain systems.
"""

import pathos.multiprocessing as pp
from os import getenv, remove
import numpy as np
from tempfile import NamedTemporaryFile
from oxDNA_analysis_tools.UTILS.readers import blocks
from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2

#actually unused these days, but just in case...
def get_n_cpu():
    """
        Gets the number of CPUs available from either the slum environment or pathos' check of available resources.

        Returns:
            available_cpus (int): Either the number of slurm tasks assigned to the process or the count of CPUs on the machine.
    """
    try:
        available_cpus = int(getenv('SLURM_NTASKS'))
    except:
        available_cpus = int(pp.cpu_count()/2)

    return available_cpus

def split_trajectory(traj_file, top_file, num_confs, n_cpus, confs_per_processor):
    """
    Splits a trajectory file into temporary files and attaches a reader to each file.

    Parameters:
        traj_file (str): Name of the trajectory file to split.  
        top_file (str): Name of the topology file associated with the trajectory. 
        num_confs (int): The number of configurations in the trajectory.  
        n_cpus (int): The number of chunks to split the trajectory into.  
        conf_per_processor (int): The number of configurations per chunk (equivalent to floor(num_confs/n_cpus))  

    Returns:
        readers (list of LorenzoReader2s): A list of readers with each one on a unique chunk of the file.
    """
    n_files = 0
    readers = []
    files = []
    rem = num_confs % n_cpus
    
    with open(traj_file, "rb") as f:
        it = blocks(f)
        chunk = next(it) # iterator producing 1 MB chunks of the trajectory
        last_conf_byte = 0

        #create a number of temporary file equal to the number of CPUs
        while n_files < n_cpus:
            out = NamedTemporaryFile(mode='w+b', delete=False)
            conf_count = 0
            
            #If there is a remainder after dividing the number of configurations by the number of CPUs
            #Add one extra configuration to the first rem files
            if n_files < rem:
                a = 1
            else:
                a = 0

            #Find successive configuration start points and write them out to the tempfiles
            while conf_count < confs_per_processor+a:
                next_conf_byte = chunk.find(b"t", last_conf_byte+1)
                if next_conf_byte == -1:
                    out.write(chunk[last_conf_byte:])
                    try:
                        chunk = next(it)
                    except: #next() throws an error if there isn't another chunk
                        break
                    last_conf_byte = 0
                else:  
                    out.write(chunk[last_conf_byte:next_conf_byte])     
                    conf_count += 1
                    last_conf_byte = next_conf_byte 

            #create a reader from the newly created trajectory chunk
            readers.append(LorenzoReader2(out.name, top_file))
            files.append(out)
            n_files += 1
        
    return(readers, files)
            

#partitions the trajectory file to the number of workers defined by n_cpus
#each worker runs the given function on its section of the trajectory
def fire_multiprocess(traj_file, top_file, function, num_confs, n_cpus, *args):
    """
    Distributes a function over a given number of processes

    Parameters:
        traj_file (str): The name of the trajectory file to analyze.
        top_file (str): The name of the topology file associated with the trajectory.
        function (function): The analysis function to be parallelized.
        num_confs (int): The number of configurations in the trajectory.
        n_cpus (int): The number of processes to launch.
        *args: The arguments for the provided function.

    Returns:
        results (list): The results from each individual processor's run.

    Note: The manner in which to concatenate the results is function-specific so should be handled in the calling module.
    """

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["pathos"])
    confs_per_processor = int(np.floor(num_confs/n_cpus))

    reader_pool = []
    processor_pool = pp.Pool(n_cpus)

    #split_starts and split_ends are around for backwards compatability with the old parallelize algorithm
    reader_pool, tmpfiles = split_trajectory(traj_file, top_file, num_confs, n_cpus, confs_per_processor)
    split_starts = [0 for  r in reader_pool]
    split_ends = [confs_per_processor for r in reader_pool]
    rem = num_confs % n_cpus
    for i in range(rem):
        split_ends[i] += 1

    #Staple everything together, send it out to the workers, and collect the results as a list
    #Functions passed to this parallelizer must have the argument order defined by the lst variable (reader, <unique args>, number of configurations total, starting conf id, number of confs for this processor)
    #This args unpacking method was added in Python 3.6, so if you have an older version of Python that's why this isn't working
    results = []
    lst = [(r, *args, num_confs, s, e) for r, s, e in zip(reader_pool, split_starts, split_ends)]
    
    #starmap allows you to have arguments that themselves are iterables
    #async because we don't actually care what order stuff finishes in.
    results = processor_pool.starmap_async(function, lst).get()
    processor_pool.close()
    for f in tmpfiles:
        f.close()
        remove(f.name)

    return(results)



#!/usr/bin/env python3

# align_trajectory.py
# Created by: Erik Poppleton
# Date: 8/26/19
# Takes a trajectory and aligns every frame to the first one and writes a new trajectory

try:
    from Bio.SVDSuperimposer import SVDSuperimposer
except:
    from bio.SVDSuperimposer import SVDSuperimposer
from sys import exit
from oxDNA_analysis_tools.UTILS.readers import ErikReader
import numpy as np
import argparse
import os

#aligner
def align_frame(ref_conf, sup, mysystem, indexes):
    #Need to get rid of fix_diffusion artifacts or SVD doesn't work
    mysystem.inbox()
    indexed_cur_conf = mysystem.positions[indexes]

    #Superimpose the configuration to the reference
    sup.set(ref_conf, indexed_cur_conf)
    sup.run()
    rot, tran = sup.get_rotran()

    #Apply rotation and translation in one step
    mysystem.positions = np.einsum('ij, ki -> kj', rot, mysystem.positions) + tran
    mysystem.a1s = np.einsum('ij, ki -> kj', rot, mysystem.a1s)
    mysystem.a3s = np.einsum('ij, ki -> kj', rot, mysystem.a3s)

    return mysystem

def main():
    #handle commandline arguments
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Aligns each frame in a trajectory to the first frame")
    parser.add_argument('traj', type=str, nargs=1, help="The trajectory file to align")
    parser.add_argument('outfile', type=str, nargs=1, help='The name of the new trajectory file to write out')
    parser.add_argument('-i', metavar='index_file', dest='index_file', nargs=1, help='Align to only a subset of particles from a space-separated list in the provided file')
    parser.add_argument('-r', metavar='reference_structure', dest='reference_structure', nargs=1, help="Align to a provided configuration instead of the first frame.")
    args = parser.parse_args()
    
    #run system checks
    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy", "Bio"])
    
    #Parse command line arguments
    traj_file = args.traj[0]
    outfile = args.outfile[0]
    sup = SVDSuperimposer()
    
    #-i will make it only run on a subset of nucleotides.
    #The index file is a space-separated list of particle IDs
    if args.index_file:
        index_file = args.index_file[0]
        with open(index_file, 'r') as f:
            indexes = f.readline().split()
            try:
                indexes = [int(i) for i in indexes]
            except:
                print("ERROR: The index file must be a space-seperated list of particles.  These can be generated using oxView by clicking the \"Download Selected Base List\" button")
    else: 
        with ErikReader(traj_file) as r:
            indexes = list(range(len(r.read().positions)))
    
    #-r will make it align to a provided .dat file instead of the first configuration
    if args.reference_structure:
        #read reference configuration
        r = ErikReader(args.reference_structure[0])
        ref = r.read()
        ref.inbox()
        r = ErikReader(traj_file)
        ref_conf = ref.positions[indexes]
        
        mysystem = align_frame(ref_conf, sup, r.read())
    
    else:
        #read the first configuration and use it as the reference configuration for the rest
        r = ErikReader(traj_file)
        mysystem = r.read()
        mysystem.inbox()
        ref_conf = mysystem.positions[indexes]
        
    #write first configuration to output file
    mysystem.write_new(outfile)
    mysystem = r.read()
    
    
    
    #Read the trajectory one configuration at a time and perform the alignment
    while mysystem != False:
        print("working on t = ", mysystem.time)
        
        mysystem = align_frame(ref_conf, sup, mysystem, indexes)
    
        mysystem.write_append(outfile)
    
        mysystem = r.read()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#superimpose.py
#Created by: Erik Poppleton
#Date: 2/27/19
#Takes two (or more) configurations and aligns all proceeding ones to the first using the svd superimposer, then spits them out as new dat files.

try:
    from Bio.SVDSuperimposer import SVDSuperimposer
except:
    from bio.SVDSuperimposer import SVDSuperimposer
from sys import exit, stderr
from oxDNA_analysis_tools.UTILS.readers import ErikReader
import numpy as np
import argparse
import os

def main():
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="superimposes one or more structures sharing a topology to a reference structure")
    parser.add_argument('reference', type=str, nargs=1, help="The reference configuration to superimpose to")
    parser.add_argument('victims', type=str, nargs='+', help="The configuraitons to superimpose on the reference")
    parser.add_argument('-i', metavar='index_file', dest='index_file', nargs=1, help='Align to only a subset of particles from a space-separated list in the provided file')
    args = parser.parse_args()


    #run system checks
    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy", "Bio"])

    #Get the reference files
    ref_dat = args.reference[0]

    #-i will make it only run on a subset of nucleotides.
    #The index file is a space-separated list of particle IDs
    if args.index_file:
        index_file = args.index_file[0]
        with open(index_file, 'r') as f:
            indexes = f.readline().split()
            try:
                indexes = [int(i) for i in indexes]
            except:
                print("ERROR: The index file must be a space-seperated list of particles.  These can be generated using oxView by clicking the \"Download Selected Base List\" button")
    else: 
        with ErikReader(ref_dat) as r:
            indexes = list(range(len(r.read().positions)))

    #Create list of configurations to superimpose
    to_sup = []
    r = ErikReader(ref_dat)
    ref = r.read()
    ref.inbox()
    ref_conf = ref.positions[indexes]
    for i in args.victims:
        r = ErikReader(i)
        sys = r.read()
        sys.inbox()
        to_sup.append(sys)

    sup = SVDSuperimposer()

    #Run the biopython superimposer on each configuration and rewrite its configuration file
    for i, sys in enumerate(to_sup):
        indexed_cur_conf = sys.positions[indexes]
        sup.set(ref_conf, indexed_cur_conf)
        sup.run()
        rot, tran = sup.get_rotran()
        sys.positions = np.einsum('ij, ki -> kj', rot, sys.positions) + tran
        sys.a1s = np.einsum('ij, ki -> kj', rot, sys.a1s)
        sys.a3s = np.einsum('ij, ki -> kj', rot, sys.a3s)
        sys.write_new("aligned{}.dat".format(i))
        print("INFO: Wrote file aligned{}.dat".format(i), file=stderr)

if __name__ == '__main__':
    main()


#!/usr/bin/env python3

from sys import stderr
try:
    from Bio.SVDSuperimposer import SVDSuperimposer
except:
    from bio.SVDSuperimposer import SVDSuperimposer
import matplotlib.pyplot as plt
from json import loads, dumps
import numpy as np
import argparse
import os
from oxDNA_analysis_tools.UTILS.readers import ErikReader, cal_confs

def compute_deviations(reader, mean_structure, indexed_mean_structure, indexes, num_confs, start=None, stop=None):
    """
        Computes RMSF of each particle from the mean structure

        Parameters:
            reader (readers.ErikReader): An active reader on the trajectory file to analyze.
            mean_structure (numpy.array): The position of each particle in the mean configuration.  A 3xN array.
            num_confs (int): The number of configurations in the reader.  
            <optional> start (int): The starting configuration ID to begin averaging at.  Used if parallel.
            <optional> stop (int): The configuration ID on which to end the averaging.  Used if parallel.

        Returns:
            deviations (list): Each entry in the list is a numpy.array of the deviations for each particle at a given time.
    """
    if stop is None:
        stop = num_confs
    else: stop = int(stop)
    if start is None:
        start = 0
    else: start = int(start)
    confid = 0

    # Use the single-value decomposition method for superimposing configurations 
    sup = SVDSuperimposer()
    deviations = []
    RMSDs = []

    mysystem = reader.read(n_skip = start)
    
    while mysystem != False and confid < stop:
        mysystem.inbox()
        # calculate alignment transform
        cur_conf = mysystem.positions
        indexed_cur_conf = cur_conf[indexes]
        sup.set(indexed_mean_structure, indexed_cur_conf)
        sup.run()
        print("Frame number:",confid, "Time:", mysystem.time, "RMSD:", sup.get_rms())
        # realign frame
        rot, tran = sup.get_rotran()
        # align structures and collect coordinates for each frame 
        # compatible with json 
        deviations.append(
           list(np.linalg.norm(np.einsum('ij, ki -> kj', rot, cur_conf) + tran - mean_structure, axis=1))
        )
        RMSDs.append(sup.get_rms()*0.8518)
        confid += 1
        mysystem = reader.read()

    return (deviations, RMSDs)

def main():
    #handle commandline arguments
    #the positional arguments for this are: 
    # 1. the mean structure from compute_mean.py in json format
    # 2. the trajectory from which to compute the deviations
    from sys import argv
    print(argv)
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Compute the RMSD of each nucleotide from the mean structure produced by compute_mean.py")
    parser.add_argument('mean_structure', type=str, nargs=1, help="The mean structure .json file from compute_mean.py")
    parser.add_argument('trajectory', type=str, nargs=1, help='the trajectory file you wish to analyze')
    parser.add_argument('-p', metavar='num_cpus', nargs=1, type=int, dest='parallel', help="(optional) How many cores to use")
    parser.add_argument('-o', '--output', metavar='output_file', nargs=1, help='The filename to save the deviations json file to')
    parser.add_argument('-i', metavar='index_file', dest='index_file', nargs=1, help='Compute mean structure of a subset of particles from a space-separated list in the provided file')
    parser.add_argument('-r', metavar='rmsd_plot', dest='rmsd_plot', nargs=1, help='The name of the file to save the RMSD plot to.')
    parser.add_argument('-d', metavar='rmsd_data', dest='rmsd_data', nargs=1, help='The name of the file to save the RNSD data in json format.')
    args = parser.parse_args()

    #system check
    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "Bio", "numpy", "matplotlib"])

    #-o names the output file
    if args.output:
        outfile = args.output[0].strip()
        if not outfile.split(".")[-1] == 'json':
            outfile += ".json"
    else: 
        outfile = "devs.json"
        print("INFO: No outfile name provided, defaulting to \"{}\"".format(outfile), file=stderr)

    #prepare the data files and calculate how many configurations there are to run
    traj_file = args.trajectory[0]
    parallel = args.parallel
    if parallel:
        from oxDNA_analysis_tools.UTILS import parallelize_erik_onefile
        n_cpus = args.parallel[0]
    num_confs = cal_confs(traj_file)

    #-i will make it only run on a subset of nucleotides.
    #The index file is a space-separated list of particle IDs
    if args.index_file:
        index_file = args.index_file[0]
        with open(index_file, 'r') as f:
            indexes = f.readline().split()
            try:
                indexes = [int(i) for i in indexes]
            except:
                print("ERROR: The index file must be a space-seperated list of particles.  These can be generated using oxView by clicking the \"Download Selected Base List\" button")
    else: 
        with ErikReader(traj_file) as r:
            indexes = list(range(len(r.read().positions)))

    #-r names the file to print the RMSD plot to
    if args.rmsd_plot:
        plot_name = args.rmsd_plot[0]
    else:
        plot_name = 'rmsd.png'

    # -d names the file to print the RMSD data to
    if args.rmsd_data:
        data_file = args.rmsd_data[0]

    # load mean structure 
    mean_structure_file = args.mean_structure[0]
    with open(mean_structure_file) as file:
        mean_data = loads(
            file.read()
        )
    mean_structure = np.array(mean_data["g_mean"])
    indexed_mean_structure = mean_structure[indexes]
    print("INFO: mean structure loaded", file=stderr)

    #Calculate deviations, in parallel if available
    if not parallel:
        print("INFO: Computing deviations from the mean of {} configurations with an alignment of {} particles using 1 core.".format(num_confs, len(indexed_mean_structure)), file=stderr)
        r = ErikReader(traj_file)
        deviations, RMSDs = compute_deviations(r, mean_structure, indexed_mean_structure, indexes, num_confs)

    #If parallel, the trajectory is split into a number of chunks equal to the number of CPUs available.
    #Each of those chunks is then calculated seperatley and the results are compiled .
    if parallel:
        print("INFO: Computing deviations from the mean of {} configurations with an alignment of {} particles using {} cores.".format(num_confs, len(indexed_mean_structure), n_cpus), file=stderr)
        deviations = []
        RMSDs = []
        out = parallelize_erik_onefile.fire_multiprocess(traj_file, compute_deviations, num_confs, n_cpus, mean_structure, indexed_mean_structure, indexes)
        [deviations.extend(i[0]) for i in out]
        [RMSDs.extend(i[1]) for i in out]

    #compute_deviations() returns the deviation of every particle in every configuration
    #take the mean of the per-configuration deviations to get the RMSF
    rmsfs = np.sqrt(np.mean(np.square(np.array(deviations)), axis=0))*0.8518

    #write the deviations to a json file
    print("INFO: writing deviations to {}".format(outfile), file=stderr)
    with open(outfile,"w") as file:
        file.write(
            dumps({
            "RMSF (nm)" : rmsfs.tolist()
        }))

    #plot RMSDs
    print("INFO: writing RMSD plot to {}".format(plot_name), file=stderr)
    plt.plot(RMSDs)
    plt.axhline(np.mean(RMSDs), color='red')
    plt.xlabel('Configuration')
    plt.ylabel('RMSD (nm)')
    plt.savefig(plot_name)

    #print RMSDs
    print("INFO: writing RMSD data to {}".format(data_file), file=stderr)
    if args.rmsd_data:
        with open(data_file, 'w') as f:
            f.write(dumps({
                "RMSD (nm)" : RMSDs
            }))

if __name__ == '__main__':
    main()
    


import argparse
import matplotlib.pyplot as plt
import numpy as np
import os
from sys import stderr

def main():
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="A very simple plotting utility for oxDNA.org")
    parser.add_argument('energy', nargs='+', help='Energy files to plot')
    parser.add_argument('-o', '--output', metavar='output_file', nargs=1, help='The name to save the graph file to')
    parser.add_argument('-f', '--format', metavar='<histogram/trajectory/both>', nargs=1, help='Output format for the graphs.  Defaults to histogram.  Options are \"histogram\", \"trajectory\", and \"both\"')
    
    args = parser.parse_args()

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy", "matplotlib"])

    #get file name
    energy_files = args.energy

    #-o names the output file
    if args.output:
        outfile = args.output[0]
    else:
        outfile = 'energy.png'

    #-f defines which type of graph to produce
    hist = False
    line = False
    if args.format:
        if "histogram" in args.format:
            hist = True
        if "trajectory" in args.format:
            line = True
        if "both" in args.format:
            hist = line = True
        if hist == line == False:
            print("ERROR: unrecognized graph format\nAccepted formats are \"histogram\", \"trajectory\", and \"both\"", file=stderr)
            exit(1)
    else:
        print("INFO: No graph format specified, defaulting to histogram", file=stderr)
        hist = True

    all_times = []
    all_energies = []
    for efile in energy_files:
        times = []
        energies = []
        with open(efile, 'r') as f:
            l = f.readline()
            while l:
                times.append(float(l.split()[0]))
                energies.append(float(l.split()[1]))
                l = f.readline()
        all_times.append(times)
        all_energies.append(energies)

    names = ["1", "2", "3"]

    if outfile and hist == True:
        if line == True:
            out = outfile[:outfile.find(".")]+"_hist"+outfile[outfile.find("."):]
        else:
            out = outfile
    
        bins = np.linspace(min([min(e) for e in all_energies]), max([max(e) for e in all_energies]), 40)
    
        artists = []
        for i,elist in enumerate(all_energies):
            a = plt.hist(elist, bins, weights=np.ones(len(elist)) / len(elist),  alpha=0.3, label=names[i], histtype=u'stepfilled', edgecolor='k')
            artists.append(a)
        plt.legend(labels=names)
        plt.xlabel("Energy per particle (SU)")
        plt.ylabel("Normalized frequency")
        if outfile:
            print("INFO: Saving histogram to {}".format(out), file=stderr)
            plt.savefig(out)
        else:
            plt.show()

    #make a trajectory plot
    if outfile and line == True:
        if hist == True:
            plt.clf()
            out = outfile[:outfile.find(".")]+"_traj"+outfile[outfile.find("."):]
        else:
            out = outfile
        
        artists = []
        for tlist,elist in zip(all_times, all_energies):
            a = plt.plot(tlist, elist, alpha = 0.5)
            artists.append(a)
        plt.legend(labels=names)
        plt.xlabel("Time (SU)")
        plt.ylabel("Energy (SU)")
        if outfile:
            print("INFO: Saving line plot to {}".format(out), file=stderr)
            plt.savefig(out)
        else:
            plt.show()
    
if __name__ == '__main__':
    main()    




#!/usr/bin/env python3

# Created by Hao Liu 
# Date 01/22/2019 
# A short script generating force file from the given .dat and .top

from sys import stderr
import os
from oxDNA_analysis_tools.output_bonds import output_bonds
import argparse
from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2, get_input_parameter

def main():
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Create an external forces file enforcing the current base-pairing arrangement")
    parser.add_argument('inputfile', type=str, nargs=1, help="The inputfile used to run the simulation")
    parser.add_argument('configuration', type=str, nargs=1, help="The configuration to generate the forces from")
    parser.add_argument('-o', '--output', type=str, nargs=1, help='name of the file to write the angle list to')
    parser.add_argument('-f', '--pairs', type=str, nargs=1, help='name of the file to write the designed pairs list to')

    args = parser.parse_args()

    #Process command line arguments:
    inputfile = args.inputfile[0]
    conf_file = args.configuration[0]

    #-o names the output file
    if args.output:
        outfile = args.output[0]
    else: 
        outfile = "forces.txt"
        print("INFO: No outfile name provided, defaulting to \"{}\"".format(outfile), file=stderr)

    if args.pairs:
        pairsfile = args.pairs[0]
    else:
        pairsfile = False     

    #Get relevant parameters from the input file
    top_file = get_input_parameter(inputfile, "topology")

    #get base pairs 
    r = LorenzoReader2(conf_file, top_file)
    mysystem = r._get_system()
    out = output_bonds(inputfile, mysystem)
    out = out.split('\n')

    #Find out the forming bonds series 
    print("INFO: Analyze the output...", file=stderr)
    Bonded = {}
    for i in out:
        if i[0] == '#':
            continue
        splitline = i.split(' ')
        try:
            HB = float(splitline[6])
        except:
            continue
        if HB < -0.001:
            ntid0 = int(splitline[0])
            ntid1 = int(splitline[1])
            if ntid0 not in Bonded:
                Bonded[ntid0] = ntid1
            if ntid1 not in Bonded:
                Bonded[ntid1] = ntid0

    lines = []
    pairlines = []
    mutual_trap_template = '{ \ntype = mutual_trap\nparticle = %d\nstiff = 0.9\nr0 = 1.2\nref_particle = %d\nPBC=1\n}\n'
    for key in sorted(Bonded):
        from_particle_id = key
        to_particle_id = Bonded[key]
        if from_particle_id < to_particle_id:
            if pairsfile:
                pairlines.append("{} {}\n".format(from_particle_id, to_particle_id))
            lines.append(mutual_trap_template % (from_particle_id,to_particle_id))
            lines.append(mutual_trap_template % (to_particle_id,from_particle_id))

    if pairsfile:
        with open(pairsfile, "w") as file:
            file.writelines(pairlines)
            print("INFO: Wrote pairs to {}".format(pairsfile), file=stderr)

    with open(outfile, "w") as file:
            file.writelines(lines)
            print("INFO: Job finished. Wrote forces to {}".format(outfile), file=stderr)

if __name__ == '__main__':
    main()




#!/usr/bin/env python

import numpy as np
from sys import exit, stderr
import argparse
import os
import matplotlib.pyplot as plt

#Calculates distance taking PBC into account
def min_image(p1, p2, box):
    p1 = p1 - (np.floor(p1/box) * box)
    p2 = p2 - (np.floor(p2/box) * box)
    diff = p1 - p2
    diff = diff - (np.round(diff/box)*box)
    return np.linalg.norm(diff)

def get_distances(trajectories, p1s, p2s):
    distances = [[] for _ in trajectories]
    for i,trajectory in enumerate(trajectories):
        with open(trajectory, 'r') as traj:
            distances[i] = [[] for _ in p1s[i]]
            l = traj.readline()
            l = traj.readline() #skip the first time line

            #get the box size
            box = np.array([l.split(' ')[2], l.split(' ')[3], l.split(' ')[4]], dtype=float)

            #Read the file line by line and make a dictionary of particle positions
            line_num = 1
            d = {}
            while l:
                #If you're at the right particle, save the COM of the particle to a dictionary
                if line_num-3 in p1s[i] or line_num-3 in p2s[i]:
                    d[line_num-3] = np.array(l.split(' ')[0:3], dtype=float)

                #if its the start of a new configuration, dump the distance data from the last one
                if 't' in l:
                    for j, (p1, p2) in enumerate(zip(p1s[i], p2s[i])):
                        p1 = d[p1]
                        p2 = d[p2]
                        distances[i][j].append(min_image(p1, p2, box)*0.85) #1 oxDNA su = 0.85 nm
                        line_num = 0
                l = traj.readline() #returns false if there's no more conf to load
                line_num += 1
            
            #catch the last configuration
            for j, (p1, p2) in enumerate(zip(p1s[i], p2s[i])):
                p1 = d[p1]
                p2 = d[p2]
                distances[i][j].append(min_image(p1, p2, box)*0.85) #1 oxDNA su = 0.85 nm

    return distances

def main():
    #handle commandline arguments
    #this program has no positional arguments, only flags
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Finds the ensemble of distances between any two particles in the system")
    parser.add_argument('-i', '--input', metavar='input', nargs='+', action='append', help='An input, trajectory, and a list of particle pairs to compare.  Can call -i multiple times to plot multiple datasets.')
    parser.add_argument('-o', '--output', metavar='output_file', nargs=1, help='The name to save the graph file to')
    parser.add_argument('-f', '--format', metavar='<histogram/trajectory/both>', nargs=1, help='Output format for the graphs.  Defaults to histogram.  Options are \"histogram\", \"trajectory\", and \"both\"')
    parser.add_argument('-d', '--data', metavar='data_file', nargs=1, help='If set, the output for the graphs will be dropped as a json to this filename for loading in oxView or your own scripts')
    parser.add_argument('-n', '--names', metavar='names', nargs='+', action='append', help='Names of the data series.  Will default to particle ids if not provided')
    parser.add_argument('-c', metavar='cluster', dest='cluster', action='store_const', const=True, default=False, help="Run the clusterer on each configuration's distance?")
    args = parser.parse_args()

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "matplotlib", "numpy"])

    #-i requires 4 or more arguments, the topology file of the structure, the trajectory to analyze, and any number of particle pairs to compute the distance between.
    try:
        input_files = [i[0] for i in args.input]
        trajectories = [i[1] for i in args.input]
        p1s = [i[2::2] for i in args.input]
        p2s = [i[3::2] for i in args.input]
        p1s = [[int(j) for j in i] for i in p1s]
        p2s = [[int(j) for j in i] for i in p2s]

    except Exception as e:
        print("ERROR:", e)
        parser.print_help()
        exit(1)
    
    #get number of distances to calculate
    n_dists = sum([len(l) for l in p1s])

    #Make sure that the input is correctly formatted
    if(len(input_files) != len(trajectories)):
        print("ERROR: bad input arguments\nPlease supply an equal number of input and, trajectory files", file=stderr)
        exit(1)
    if len(p1s) != len(p2s):
        print("ERROR: bad input arguments\nPlease supply an even number of particles", file=stderr)
        exit(1)


    #-o names the output file
    if args.output:
        outfile = args.output[0]
    else: 
        print("INFO: No outfile name provided, defaulting to \"distance.png\"", file=stderr)
        outfile = "distance.png"

    #-f defines which type of graph to produce
    hist = False
    lineplt = False
    if args.format:
        if "histogram" in args.format:
            hist = True
        if "trajectory" in args.format:
            lineplt = True
        if "both" in args.format:
            hist = True
            lineplt = True
        if hist == lineplt == False:
            print("ERROR: unrecognized graph format\nAccepted formats are \"histogram\", \"trajectory\", and \"both\"", file=stderr)
            exit(1)
    else:
        print("INFO: No graph format specified, defaulting to histogram", file=stderr)
        hist = True

    #-c makes it run the clusterer on the output
    cluster = args.cluster
    
    #get the specified distances
    distances = get_distances(trajectories, p1s, p2s)

    # -n sets the names of the data series
    if args.names:
        names = args.names[0]
        if len(names) < n_dists:
            print("WARNING: Names list too short.  There are {} items in names and {} distances were calculated.  Will pad with particle IDs".format(len(names), n_dists), file=stderr)
            for i in range(len(names), len(distances)):
                names.append("{}-{}".format([j for sl in p1s for j in sl][i], [j for sl in p2s for j in sl][i]))
        if len(names) > n_dists:
            print("WARNING: Names list too long. There are {} items in names and {} distances were calculated.  Truncating to be the same as distances".format(len(names), n_dists), file=stderr)
            names = names[:n_dists]

    else:
        print("INFO: Defaulting to particle IDs as data series names")
        names = ["{}-{}".format(p1, p2) for p1, p2 in zip([i for sl in p1s for i in sl], [i for sl in p2s for i in sl])]
    
    # -d will dump the distances as json files for loading with the trajectories in oxView
    if args.data:
        from json import dump
        if len(trajectories) > 1:
            print("INFO: distance lists from separate trajectories are printed to separate files for oxView compatibility.  Trajectory numbers will be appended to your provided data file name.", file=stderr)
            file_names = ["{}_{}.json".format(args.data[0].strip('.json'), i) for i,_ in enumerate(trajectories)]
        else:
            file_names = [args.data[0].strip('.json')+'.json']
        names_by_traj = [['{}-{}'.format(p1, p2) for p1, p2 in zip(p1l, p2l)] for p1l, p2l in zip(p1s, p2s)]
        
        for file_name, ns, dist_list in zip(file_names, names_by_traj, distances):
            obj = {}
            for n, d in zip(ns, dist_list):
                obj[n] = d        
            with open(file_name, 'w+') as f:
                print("INFO: writing data to {}.  This can be opened in oxView using the Order parameter selector".format(file_name))
                dump(obj, f)

    #convert the distance list into numpy arrays because they're easier to work with
    for i, l in enumerate(distances):
        distances[i] = np.array(l)
    
    means = [np.mean(i, axis=1) for i in distances]
    medians = [np.median(i, axis=1) for i in distances]
    stdevs = [np.std(i, axis=1) for i in distances]

    #get some min/max values to make the plots pretty
    lower = min((l.min() for l in distances))
    upper = max((l.max() for l in distances))

    #those horrific list comprehensions unpack lists of lists into a single list
    print("input:\t", end='')
    [print("{}-{}\t".format(p1, p2), end='') for p1, p2 in zip([i for sl in p1s for i in sl], [i for sl in p2s for i in sl])]
    print("")

    print("name:\t", end='')
    [print("{}\t".format(t), end='') for t in names[:n_dists]]
    print("")

    print("mean:\t", end='')
    [print("{:.2f}\t".format(m), end='') for m in [i for sl in means for i in sl]]
    print("")

    print("stdev:\t", end='')
    [print("{:.2f}\t".format(s), end='') for s in [i for sl in stdevs for i in sl]]
    print("")

    print("median:\t", end='')
    [print("{:.2f}\t".format(m), end='') for m in [i for sl in medians for i in sl]]
    print("")

    #make a histogram
    if hist == True:
        if lineplt == True:
            #if making two plots, automatically append the plot type to the output file name
            out = outfile[:outfile.find(".")]+"_hist"+outfile[outfile.find("."):]
        else:
            out = outfile
        bins = np.linspace(np.floor(lower-(lower*0.1)), np.ceil(upper+(upper*0.1)), 60)
        graph_count = 0
        for traj_set in distances:
            for dist_list in traj_set:
                a = plt.hist(dist_list, bins, weights=np.ones(len(dist_list)) / len(dist_list),  alpha=0.5, histtype=u'stepfilled', edgecolor='k', label=names[graph_count])
                graph_count += 1
        plt.xlabel("Distance (nm)")
        plt.ylabel("Normalized frequency")
        plt.legend()
        #plt.show()
        print("INFO: Writing histogram to file {}".format(out), file=stderr)
        plt.savefig("{}".format(out))

    #make a trajectory plot
    if lineplt == True:
        if hist == True:
            #clear the histogram plot
            plt.clf()
            #if making two plots, automatically append the plot type to the output file name
            out = outfile[:outfile.find(".")]+"_traj"+outfile[outfile.find("."):]
        else:
            out = outfile
        graph_count = 0
        for traj_set in distances:
            for dist_list in traj_set:
                a = plt.plot(dist_list, alpha=0.5, label=names[graph_count])
                graph_count += 1
        plt.xlabel("Simulation Steps")
        plt.ylabel("Distance (nm)")
        plt.legend()
        #plt.show()
        print("INFO: Writing trajectory plot to file {}".format(out), file=stderr)
        plt.savefig("{}".format(out))

    if cluster == True:
        if not all([x == trajectories[0] for x in trajectories]):
            print("ERROR: Clustering can only be run on a single trajectory", file=stderr)
            exit(1)
        from oxDNA_analysis_tools.clustering import perform_DBSCAN

        labs = perform_DBSCAN(distances[0].T, len(distances[0][0]), trajectories[0], input_files[0], "euclidean")

if __name__ == '__main__':
    main()
    


# This script sets up the command line entry points to create the namespace "oat"

# Created by Erik Poppleton, October 2021

def main():
    from sys import argv
    from importlib import import_module

    try:
        script = argv.pop(1)
    except:
        from os import listdir, path
        print("Please select an analysis to run.  The available options are:")
        contents =listdir(path.dirname(__file__))

        #list the scripts that aren't __init__.py and entry.py
        contents = [c.replace('.py', '') for c in contents if ('.py' in c and not '__' in c and not path.basename(__file__) in c)]
        contents.sort()
        print('\n'.join(contents))
        exit(1)

    to_call = import_module('oxDNA_analysis_tools.{}'.format(script))

    to_call.main()


#!/usr/bin/env python3

from os import environ, path
import sys
import subprocess
import tempfile
import numpy as np

from oxDNA_analysis_tools.config import set_analysis_path
PROCESSPROGRAM = set_analysis_path()

from oxDNA_analysis_tools.config import check_dependencies
check_dependencies(["numpy"])

def contact_map (inputfile, mysystem, return_full_matrix):
    """
    Computes the distance between every pair of nucleotides and creates a matrix of these distances.

    Parameters:
        inputfile (string): the input file with which the simulation was run,
        mysystem (base.System): a base.py system object containing the system to analyze.
        return_full_matrix (bool): The matrix is symmetric. Return only the lower half or the whole matrix?
    
    Returns:
        distances (numpy.array): The matrix containing pairwise distances between every pair of nucleotides.
    """
    tempfile_obj = tempfile.NamedTemporaryFile()

    #the algorithm to find the distances is written into an oxDNA observable for improved speed on large systems.
    command_for_data =  'analysis_data_output_1 = { \n name = stdout \n print_every = 1 \n col_1 = { \n type=contact_map \n } \n}'
    launchargs = [PROCESSPROGRAM, inputfile ,'trajectory_file='+tempfile_obj.name,command_for_data]
    n = mysystem.N

    #write the system to a tempfile so it can be fed to oxDNA in a subprocess
    mysystem.print_lorenzo_output(tempfile_obj.name,'/dev/null')
    tempfile_obj.flush()
    process = subprocess.run(launchargs,stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    
    #process output
    out = process.stdout.strip()
    err = process.stderr.strip()
    for line in err.split("\n"):
        if "CRITICAL" in line or "ERROR" in line or "Error" in line:
           print (err)
           sys.exit(1)
    distances = np.array(out.split(' '), dtype=float)

    #the distance matrix is diagonal, so the observable only calculates half in a compressed format
    #if the full matrix is needed, this copies it out.
    if return_full_matrix:
        full_matrix = np.empty((n, n))
        for i in range(n):
            for j in range(n):
                if i==j:
                    full_matrix[i][j] = 0
                elif j > i:
                    full_matrix[i][j] = distances[int(i*n - ((i*i)+i)*0.5 + (j-i-1))]
                elif i > j:
                    full_matrix[i][j] = distances[int(j*n - ((j*j)+j)*0.5 + (i-j-1))]
        return(full_matrix)
    else:
        return(distances)

def main():
    import argparse
    import matplotlib.pyplot as plt
    from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2, get_input_parameter

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy", "matplotlib"])

    #get commandline arguments
    parser = argparse.ArgumentParser(prog = path.basename(__file__), description="Calculate and display the contact map for a structure")
    parser.add_argument('inputfile', type=str, nargs=1, help="The inputfile used to run the simulation")
    parser.add_argument('trajectory', type=str, nargs=1, help="The file containing the configurations of which the contact map is needed")
    parser.add_argument('-v', dest='visualize', action='store_const', const=True, default=False, help="should we display the contact map once its calculated? Only recommend if there are few confs.")

    args = parser.parse_args()
    visualize = args.visualize
    inputfile = args.inputfile[0]
    traj_file = args.trajectory[0]

    #process files
    top_file = get_input_parameter(inputfile, "topology")
    if "RNA" in get_input_parameter(inputfile, "interaction_type"):
        environ["OXRNA"] = "1"
    else:
        environ["OXRNA"] = "0"

    #create system object from first configuration in the trajectory
    r = LorenzoReader2(traj_file, top_file)
    system = r._get_system()

    #for every configuration, create a graphical contact map
    while system:
        m = contact_map(inputfile, system, True)
        if visualize:
            fig, ax = plt.subplots()
            a = ax.imshow(m, cmap='viridis', origin='lower')
            ax.set(title = "interaction network",
                ylabel="nucleotide id",
                xlabel="nucleotide id")
            b = fig.colorbar(a, ax=ax)
            b.set_label("distance", rotation = 270)
            plt.show()
        system = r._get_system()

if __name__ == '__main__':
    main()
    



#!/usr/bin/env python
# from original oxDNA repository
#A utility that prints out the number of hydrogen bonds between different strands in the system

import numpy as np
from os import environ, path, getcwd
from sys import stderr, exit
from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2, get_input_parameter
import subprocess
import tempfile

from oxDNA_analysis_tools.config import set_analysis_path
PROCESSPROGRAM = set_analysis_path()

command_for_data =  'analysis_data_output_1={ \n name = stdout \n print_every = 1 \n col_1 = { \n type=pair_energy \n} \n}'

def output_bonds (inputfile, system):
	tempfile_obj = tempfile.NamedTemporaryFile()
	temp_top = tempfile.NamedTemporaryFile()
	system.print_lorenzo_output(tempfile_obj.name,temp_top.name)
	tempfile_obj.flush()

	launchargs = [PROCESSPROGRAM, inputfile, 'trajectory_file='+tempfile_obj.name, 'topology='+temp_top.name, command_for_data]

	myinput = subprocess.run(launchargs,stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
	out = myinput.stdout.strip()
	err = myinput.stderr.strip()
	for line in err.split('\n'):
		if "CRITICAL" in line or "ERROR" in line:
			print(err, file=stderr)
			print("Bad configuration found at t = {}, outputting blank string".format(int(system._time)), file=stderr)
			return ''
	return out

def main():
	import argparse
	parser = argparse.ArgumentParser(prog = path.basename(__file__), description="List all the interactions between nucleotides")
	parser.add_argument('inputfile', type=str, nargs=1, help="The inputfile used to run the simulation")
	parser.add_argument('trajectory', type=str, nargs=1, help='the trajectory file you wish to analyze')
	parser.add_argument('-v', type=str, nargs=1, dest='outfile', help='if you want instead average per-particle energy as a viewer JSON')
	args = parser.parse_args()

	from oxDNA_analysis_tools.config import check_dependencies
	check_dependencies(["python", "numpy"])

	traj_file = args.trajectory[0]
	inputfile = args.inputfile[0]

	try:
		outfile = args.outfile[0]
		visualize = True
	except:
		visualize = False

	if path.dirname(inputfile) != getcwd():
		sim_directory = path.dirname(inputfile)
	else:
		sim_directory = ""

	top_file = sim_directory + get_input_parameter(inputfile, "topology")
	if "RNA" in get_input_parameter(inputfile, "interaction_type"):
		environ["OXRNA"] = "1"
	else:
		environ["OXRNA"] = "0"
	import oxDNA_analysis_tools.UTILS.base #this needs to be imported after the model type is set

	myreader = LorenzoReader2(traj_file,top_file)
	mysystem = myreader._get_system()

	energies = np.zeros(mysystem.N)
	count = 0

	while mysystem != False:
		out = output_bonds(inputfile, mysystem)
		if visualize:
			for line in out.split('\n'):
				if not (line.startswith('#') or line == ''):
					line = [float(l) for l in line.split(' ')]
					energies[int(line[0])] += sum(line[2:])
					energies[int(line[1])] += sum(line[2:])
		else:
			print(out)

		count += 1
		mysystem = myreader._get_system()

	if visualize:
		energies *= (41.42/count)
		with open(outfile, "w+") as file:
			file.write("{\n\"Energy (pN nm)\" : [")
			file.write(str(energies[0]))
			for n in energies[1:]:
				file.write(", {}".format(n))
			file.write("] \n}")

if __name__ == '__main__':
    main()





#see: https://dna.physics.ox.ac.uk/index.php/Documentation#External_Forces

"""
A spring force that pulls a particle towards the position of another particle

Parameters:
    particle (int): the particle that the force acts upon
    ref_particle (int): the particle that the particle will be pulled towards
    stiff (float): the force constant of the spring (in simulation units)
    r0 (float): the equlibrium distance of the spring
    PBC (bool): does the force calculation take PBC into account (almost always 1)
"""
def mutual_trap(particle, ref_particle, stiff, r0, PBC):
    PBC = int(PBC)
    return({
        "type" : "mutual_trap",
        "particle" : particle,
        "ref_particle" : ref_particle,
        "stiff" : stiff, 
        "r0" : r0,
        "PBC" : PBC
    })

"""
A linear force along a vector

Parameters:
    particle (int): the particle that the force acts upon
    f0 (float): the initial strength of the force at t=0 (in simulation units)
    rate (float or SN string): growing rate of the force (simulation units/timestep)
    dir ([float, float, float]): the direction of the force
"""
def string(particle, f0, rate, direction):
    return({
        "type" : "string",
        "particle" : particle, 
        "f0" : f0, 
        "rate" : rate, 
        "dir" : direction 
    })

"""
A linear potential well that traps a particle

Parameters:
    particle (int): the particle that the force acts upon
    pos0 ([float, float, float]): the position of the trap at t=0
    stiff (float): the stiffness of the trap (force = stiff * dx)
    rate (float): the velocity of the trap (simulation units/time step)
    direction ([float, float, float]): the direction of movement of the trap
"""
def harmonic_trap(particle, pos0, stiff, rate, direction):
    return({
        "type" : "trap",
        "particle" : particle, 
        "pos0" : pos0,
        "rate" : rate,
        "dir" : direction
    })

"""
A harmonic trap that rotates in space with constant angular velocity

Parameters:
    particle (int): the particle that the force acts upon
    pos0 ([float, float, float]): the position of the trap at t=0
    stiff (float): the stiffness of the trap (force = stiff * dx)
    rate (float): the angular velocity of the trap (simulation units/time step)
    base (float): initial phase of the trap
    axis ([float, float, float]): the rotation axis of the trap
    mask([float, float, float]): the masking vector of the trap (force vector is element-wise multiplied by mask)
"""
def rotating_harmonic_trap(particle, stiff, rate, base, pos0, center, axis, mask):
    return({
        "type" : "twist", 
        "particle" : particle,
        "stiff" : stiff,
        "rate" : rate,
        "base" : base,
        "pos0" : pos0,
        "center" : center,
        "axis" : center,
        "mask" : mask
    })

"""
A plane that forces the affected particle to stay on one side.

Parameters:
    particle (int): the particle that the force acts upon.  -1 will act on whole system.
    stiff (float): the stiffness of the trap (force = stiff * distance below plane)
    dir ([float, float, float]): the normal vecor to the plane
    position(float): position of the plane (plane is d0*x + d1*y + d2*z + position = 0)
"""
def repulsion_plane(particle, stiff, direction, position):
    return({
        "type" : "repulsion_plane",
        "particle" : particle,
        "stiff" : stiff,
        "dir" : direction,
        "position" : position
    })

"""
A sphere that encloses the particle

Parameters:
    particle (int): the particle that the force acts upon
    center ([float, float, float]): the center of the sphere
    stiff (float): stiffness of trap
    r0 (float): radius of sphere at t=0
    rate (float): the sphere's radius changes to r = r0 + rate*t
"""
def repulsion_sphere(particle, center, stiff, r0, rate=1):
    return({
        "type" : "sphere",
        "center" : center,
        "stiff" : stiff,
        "r0" : r0,
        "rate" : rate
    })



"""
This parallelizer attaches multiple ErikReaders to a single trajectory.
This is less memory-intensive than multi-file, but is a significant performance decrease on certain systems.
"""

import pathos.multiprocessing as pp
from os import getenv
import numpy as np
from oxDNA_analysis_tools.UTILS.readers import ErikReader

#actually unused these days, but just in case...
def get_n_cpu():
    try:
        available_cpus = int(getenv('SLURM_NTASKS'))
    except:
        available_cpus = int(pp.cpu_count()/2)

    return available_cpus

#partitions the trajectory file to the number of workers defined by n_cpus
#each worker runs the given function on its section of the trajectory
def fire_multiprocess(traj_file, function, num_confs, n_cpus, *args, **kwargs):
    confs_per_processor = int(np.floor(num_confs/n_cpus))

    reader_pool = []
    processor_pool = pp.Pool(n_cpus)

    #for calculations on symmetric matricies (eRMSD)
    #can't just hand each line to the parallelizer
    if ("matrix", True) in kwargs.items():
        total_calculations = sum([(num_confs-i) for i in range(1, num_confs)])
        calcs_per_cpus = total_calculations/n_cpus
        split_ends = []
        i = 0
        while i < num_confs:
            e = 0
            calcs = 0
            while calcs < calcs_per_cpus:
                calcs += num_confs-i
                e += 1
                i += 1
                if i >=num_confs: break
            split_ends.append(e)
    
    #define sizes of trajectory chunks
    else:
        split_ends = [confs_per_processor for _ in range(n_cpus)]
        split_ends[-1] += num_confs%n_cpus #last chunk gets all the leftovers
    
    #now figure out which configuration each chunk starts on
    split_starts = [0]
    for i in range(n_cpus):
        reader_pool.append(ErikReader(traj_file))
        #rint(split_starts[i-1], split_ends[i-1])
        if i!= 0:
            split_starts.append(split_starts[i-1]+split_ends[i-1])

    #staple everything together, send it out to the workers, and collect the results as a list
    results = []
    lst = [(r, *args, num_confs, s, e) for r, s, e in zip(reader_pool, split_starts, split_ends)]
    results = processor_pool.starmap_async(function, lst).get()
    processor_pool.close()

    return(results)



#!/usr/bin/env python3

#A contact-map based clustering and flexibility quantification algorithm
#Created by: Erik Poppleton
#Date: 11/13/18
#Python2
#Takes topology and trajectory and calculates contact mapping between all nucleotides and calculates deviation score
#between snapshots

import numpy as np
from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2, cal_confs, get_input_parameter
from os import environ, path
import subprocess
import pickle
import matplotlib.pyplot as plt
from oxDNA_analysis_tools.UTILS.all_vectors import all_vectors
import argparse
from oxDNA_analysis_tools.clustering import perform_DBSCAN
from oxDNA_analysis_tools.UTILS import parallelize_lorenzo_onefile

#a matrix of vectors in local cylindrical coordinates
def calc_matrix(system, inputfile):
    interaction_sites = np.empty([len(system._nucleotides),4, 3]) #length of nucleotide list, 3-element position, 3-element a1 vector, 3-element a2 vector 3-element a3 vector
    for i,n in enumerate(system._nucleotides):
        interaction_sites[i] = np.array([n.cm_pos, n._a1, n._a2, n._a3]) #note that this uses the CMpos, not the interaction site
        #Because I don't know whether you're feeding me DNA or RNA without asking

    #calculate the interaction network    
    #from warnings import catch_warnings, simplefilter
    Vvec = all_vectors(inputfile, system, True)
    Vtransform_matrix = np.linalg.inv(np.array(interaction_sites[:, 1:].transpose(0, 2, 1))) 
    Vvec = np.einsum('ijk,ilk -> ilj', Vtransform_matrix, Vvec) #magic to apply the transform matrix to each vector in Vvec
    #in the bussi paper, they scale the structures to make the stacking and H-bonding the same length
    #I'm not going to do that to save us a computation, also because oxDNA has very similar H-bonding
    #and stacking lengths.  Using energies would probably be a more accurate method to get differences.
    #Vvec = Vvec*np.array([])
    #in the paper they also convert to cyllindrical coordinates to make figure 1.  To reproduce that I also convert to cyllindrical coords, but its not necessary for the eRMSD calculation
    #Vnetwork_matrix[:, :, 0] = np.linalg.norm(np.array([Vvec[:, :, 0], Vvec[:, :, 1], 0])) #magnitude of a1-a2 projection
    #with catch_warnings(): #dividing the whole matrix produces a divide by 0 warning on the diagonal, which is annoying
    #    simplefilter("ignore")
    #    Vnetwork_matrix[:, :, 1] = np.arctan(Vvec[:, :, 1]/Vvec[:, :, 0]) #angle of a1-a2 projection
    #Vnetwork_matrix[:, :, 2] = Vvec[:, :, 2] #a3 projection
    
    #Leaving this in here for posterity, it does the exact same thing as the above block, but ~80x slower
    '''start = time.clock()
    for i,val1 in enumerate(interaction_sites): #reference base
        transform_matrix = np.array([val1[1], val1[2], val1[3]]).T
        for j,val2 in enumerate(interaction_sites): #other base
            vec = val2[0] - val1[0]
            #if i == 0 and j == 1: print transform_matrix
            if i != j:
                transformed_vec = np.linalg.solve(transform_matrix, vec)
                proj = [transformed_vec[0], transformed_vec[1], 0]
                p = np.linalg.norm(proj)
                theta = np.arctan((transformed_vec[1] / transformed_vec[0]))
                z = transformed_vec[2]
            else: 
                theta = 0.
                p = 0.
                z = 0.
            network_matrix[i][j] = np.array([p, theta, z]) #cylindrical coords
    print network_matrix[0, 1]
    print time.clock() - start'''
    return(Vvec)

#part of the eRMSD calculation
def calc_gvec(r):
    #if we actually used cyllindrical coords, this line would be necessary
    #r = np.array([connection[0] * np.cos(connection[1]), connection[0] * np.sin(connection[1]), connection[2]]) #this is cartesian wrt the a-vectors
    r_norm = np.linalg.norm(r)
    y = 2.9 #cutoff distance
    if r_norm < y and r_norm != 0:
        G = np.array([np.sin(y * r_norm) * (r[0]/r_norm), np.sin(y * r_norm) * (r[1]/r_norm), np.sin(y * r_norm) * (r[2]/r_norm), 1+np.cos(y * r_norm)])
    else:
        G = np.array([0., 0., 0., 0.])
    return G

#calculate eRMSD, for detailed information see: 
# Bottaro, S., Di Palma, F., & Bussi, G. (2014). 
# The role of nucleobase interactions in RNA structure and dynamics. 
def calc_eRMSD(mat1, mat2):
    SD = 0
    for ie, i in enumerate(mat1):
        for je,  j in enumerate(i):
            G1 = calc_gvec(j)
            G2 = calc_gvec(mat2[ie][je])
            SD += np.square(np.linalg.norm(G1-G2))
    SD /= len(i) #now its MSD
    SD = np.sqrt(SD) #and now its RMSD
    return SD

#reads in every combination of configurations and calls calc_eRMSD()
#only fills in half the matrix since eRMSDs are symmetric, though the connection matrix is not
def get_eRMSDs(r1, r2, inputfile, traj_file, top_file, num_confs, start=None, stop=None):
    if stop is None:
        stop = num_confs
    else: stop = int(stop)
    if start is None:
        start = 0
    else: start = int(start)
    confid = 0

    system1 = r1._get_system(N_skip=start)
    system2 = r2._get_system(N_skip=start+1)
    eRMSDs = np.zeros((num_confs, num_confs))
    i = start
    j = start+1
    while system1 != False and confid < stop: 
        print ("working on configuration", i, "t =", system1._time)
        system1.inbox()
        mat1 = calc_matrix(system1, inputfile)
        while system2:
            print ("working on configuration", i, "compared to", j)
            system2.inbox()
            mat2 = calc_matrix(system2, inputfile)
            eRMSDs[i][j] = calc_eRMSD(mat1, mat2)
            system2 = r2._get_system()
            j+=1

        i += 1
        j = i+1
        confid += 1
        system1 = r1._get_system()
        r2 = LorenzoReader2(traj_file, top_file)
        system2 = r2._get_system(N_skip=j)

    return(eRMSDs)

##############################################################################################################
#Welcome to the meat
#First, we read arguments and calculate the eRMSD between all structures
#This will not produce any outputs
def main():
    parser = argparse.ArgumentParser(prog = path.basename(__file__), description="Calculate differences between structures and automatically apply DBSCAN to retrieve clusters")
    parser.add_argument('inputfile', type=str, nargs=1, help="The inputfile used to run the simulation")
    parser.add_argument('trajectory', type=str, nargs=1, help='the trajectory file you wish to analyze')
    parser.add_argument('-p', metavar='num_cpus', nargs=1, type=int, dest='parallel', help="(optional) How many cores to use")
    args = parser.parse_args()

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy", "matplotlib"])

    traj_file = args.trajectory[0]
    inputfile = args.inputfile[0] 
    parallel = args.parallel
    if parallel:
        n_cpus = args.parallel[0]

    top_file = get_input_parameter(inputfile, "topology")
    if "RNA" in get_input_parameter(inputfile, "interaction_type"):
        environ["OXRNA"] = "1"
    else:
        environ["OXRNA"] = "0"
    num_confs = cal_confs(traj_file)
    import UTILS.base #this needs to be imported after the model type is set


    r2 = LorenzoReader2(traj_file, top_file)

    #how do you want to get your eRMSDs?  Do you need to do the time-consuming calculation or is it done and you have a pickle?
    if not parallel:
        r1 = LorenzoReader2(traj_file, top_file)

        eRMSDs = get_eRMSDs(r1, r2, inputfile, traj_file, top_file, num_confs)
    if parallel:
        out = parallelize_lorenzo_onefile.fire_multiprocess(traj_file, top_file, get_eRMSDs, num_confs, n_cpus, r2, inputfile, traj_file, top_file, matrix=True)
        eRMSDs = np.sum((i for i in out), axis=0)
    #eRMSDs = pickle.load(open('tmp_eRMSDs', 'rb'))

    #the eRMSD matrix is actually only half a matrix
    for ni,i in enumerate(eRMSDs):
        for nj,j in enumerate(i):
            eRMSDs[nj][ni] = j
            if ni == nj:
                eRMSDs[ni][nj] = 0

    #since calculating the eRMSDs are so time-consuming to calculate we're gonna pickle it to iterate the DBSCAN later.
    with open("tmp_eRMSDs", "wb") as file:
        pickle.dump(eRMSDs, file)

    ###############################################################################################################
    #Next, we're going to perform a DBSCAN on that matrix of eRMSDs to find clusters of similar structures
    perform_DBSCAN(eRMSDs, num_confs, traj_file, inputfile, "precomputed")

    ##############################################################################################################

if __name__ == '__main__':
    main()
    